/* AbstractTrack.java

		Abstract track used in constructing tracks.

 Created by  Mike Ronan  July '00
 Modified by

 */

package hep.lcd.recon.ztracking;

import hep.lcd.event.*;

import hep.physics.*;    // VecOp's

import java.io.Serializable;
import java.util.*;


public abstract class AbstractTrack implements Track, Serializable
{
    Vector hits = new Vector();
    double p = 0., px = 0., py = 0., pz = 0.;
    int charge = 0;

    double BField=0.;
    TrackParameters parameters = null;
    TrackExtrapolations extrapolations = null;

    MCParticle mcp = null;
    double mcpFraction, mcpTrackerFraction;

    boolean fitted = false;
    double Chi2 = 0.;
    double NDF = 0;
    double errmatrix[][] = null;


    /** Create an abstract track */
    public AbstractTrack()
    {
    }

    /** Create an abstract track with associated Monte Carlo particle */
    public AbstractTrack(MCParticle mcp)
    {
	this.mcp = mcp;
    }


    /** Get charge of track */
    public int getCharge()
    {
	return parameters.omega>0 ? +1 : -1;
    }

    /** Get origin of track */
    public double[] getOrigin()
    {
	double[] result = { getOriginX(), getOriginY(), getOriginZ() };
	return result;
    }
    public double getOriginX()
    {
	int charge = parameters.omega>0 ? +1 : -1;
	//return parameters.d0 * Math.cos(parameters.phi0 + charge * halfpi);  // Wrong! d0 is signed.
	return -1. * parameters.d0 * Math.sin(parameters.phi0);
    }
    public double getOriginY()
    {
	int charge = parameters.omega>0 ? +1 : -1;
	//return parameters.d0 * Math.sin(parameters.phi0 + charge * halfpi);  // Wrong! d0 is signed.
	return parameters.d0 * Math.cos(parameters.phi0);
    }
    public double getOriginZ()
    {
	return parameters.z0;
    }

    /** Get momentum of track */
    public double getMom()
    {
	double Px = getMomentumX();
	double Py = getMomentumY();
	double Pz = getMomentumZ();
	return Math.sqrt(Px*Px + Py*Py + Pz*Pz);
    }
    /** Get momentum vector of track */
    public double[] getMomentum()
    {
	double[] result = { getMomentumX(), getMomentumY(), getMomentumZ() };
	return result;
    }
    public double getMomentumX()
    {
	return getPX();
    }
    public double getMomentumY()
    {
	return getPY();
    }
    public double getMomentumZ()
    {
	return getPZ();
    }

    public double getPX()
    {
	double pt = BField*0.3*.01/Math.abs(parameters.omega);
	return pt*Math.cos(parameters.phi0);
    }
    public double getPY()
    {
	double pt = BField*0.3*.01/Math.abs(parameters.omega);
	return pt*Math.sin(parameters.phi0);
    }
    public double getPZ()
    {
	double pt = BField*0.3*.01/Math.abs(parameters.omega);
	return pt*parameters.s;
    }


    /** Get number of track hits */
    public int getNHits() { return hits.size(); }
    /** Get track hits */
    public Vector getHits() { return hits; }
    /** Add hits on track */
    public void addHit(TrackerHit hit) { hits.addElement(hit); }


    /** Returns an array of parameters for the track. */
    public double[] getTrackParameters()
    {
	return parameters.getParameters();
    }
    /** Returns the selected parameter for the track. */
    public double getTrackParameter(int i)
    {
	return parameters.getParameter(i);
    }

    /** Returns status of track fitting. */
    public boolean fitSuccess()
    {
	return fitted;
    }
    /** Get overall chi-squared for the selected track. */
    public double getChi2()
    {
	return Chi2;
    }
    /** Get number of degrees of freedom for the track. */
    public double getNDF()
    {
	return NDF;
    }
    /** Get track error matrix. */
    public double[][] getErrorMatrix()
    {
	return errmatrix;
    }
    /** Get track error matrix element. */
    public double getErrorMatrixElement(int i, int j)
    {
	return errmatrix == null ? 0 : errmatrix[i][j];
    }


    /** Return associated Monte Carlo particle */
    public MCParticle getMCParticle() { return mcp; }
    /** Get fraction of hits from parent MC particle. */
    public double getMCPFraction() { return mcpFraction; }
    /** Get fraction of Tracker hits from parent MC particle. */
    public double getMCPTrackerFraction() { return mcpTrackerFraction; }


    /** Returns extrapolations of track to various points in the detector. */
    public TrackExtrapolations getTrackExtrapolations() { return extrapolations; }
    public boolean   reachesRadius(double r) { return extrapolations.reachesRadius(r); }
    public boolean   exitsEnd(double z) { return extrapolations.exitsEnd(z); }

    public double[] get_EM_Extrapolation()  { return extrapolations.get_EM_Extrapolation(); }
    public double[] get_HAD_Extrapolation() { return extrapolations.get_HAD_Extrapolation(); }

    public double[] get_EM_Extrapolation(double r, double z)
    {
	double[] result = new double[3];
	double[] extrap1 = extrapolations.get_EM_Extrapolation1(); double r1 = Math.sqrt(extrap1[0]*extrap1[0]+extrap1[1]*extrap1[1]);
	double[] extrap2 = extrapolations.get_EM_Extrapolation2(); double r2 = Math.sqrt(extrap2[0]*extrap2[0]+extrap2[1]*extrap2[1]);
	if (r1<r && r<r2) {
	    double fraction = (r - r1) / (r2 - r1);
	    for (int i=0;i<3;i++) {
		result[i] = extrap1[i] + fraction * (extrap2[i] - extrap1[i]);
	    }
	    if (firstEM_Barrel_Extrapolation) {
		System.out.println("   For cluster at r,z = "+r+","+z);
		// System.out.println("    and search track extrapolations x,y(,r),z of: \n"+
		// 		   " \t\t"+extrap1[0]+","+extrap1[1]+"("+r1+"),"+extrap1[2]+"\n"+
		// 		   " \t\t"+extrap2[0]+","+extrap2[1]+"("+r2+"),"+extrap2[2]);
		System.out.println("    best track extrapolation is "+result[0]+","+result[1]+"("+
				   Math.sqrt(result[0]*result[0]+result[1]*result[1])+"),"+result[2]);
		firstEM_Barrel_Extrapolation = false;
	    }
	}
	else
	if (Math.abs(extrap1[2])<Math.abs(z) && Math.abs(z)<Math.abs(extrap2[2])) {
	    double fraction = (z - extrap1[2]) / (extrap2[2] - extrap1[2]);
	    for (int i=0;i<3;i++) {
		result[i] = extrap1[i] + fraction * (extrap2[i] - extrap1[i]);
	    }
	    if (firstEM_Endcap_Extrapolation) {
		System.out.println("   For cluster at r,z = "+r+","+z);
		// System.out.println("    and search track extrapolations x,y(,r),z of: \n"+
		// 		   " \t\t"+extrap1[0]+","+extrap1[1]+"("+r1+"),"+extrap1[2]+"\n"+
		// 		   " \t\t"+extrap2[0]+","+extrap2[1]+"("+r2+"),"+extrap2[2]);
		System.out.println("    best track extrapolation is "+result[0]+","+result[1]+"("+
				   Math.sqrt(result[0]*result[0]+result[1]*result[1])+"),"+result[2]);
		firstEM_Endcap_Extrapolation = false;
	    }
	}
	else {
	    if (firstEM_Extrapolation_error) {
		System.out.println(" *** Error *** For cluster at r,z = "+r+","+z);
		System.out.println("  Track extrapolations x,y(,r),z are: \n"+
				   " \t\t"+extrap1[0]+","+extrap1[1]+"("+r1+"),"+extrap1[2]+"\n"+
				   " \t\t"+extrap2[0]+","+extrap2[1]+"("+r2+"),"+extrap2[2]);
		firstEM_Extrapolation_error = false;
	    }
	    for (int i=0;i<3;i++) {
		result[i] = extrap1[i];
	    }
	}
	return result;
    }

    public double[] get_HAD_Extrapolation(double r, double z)
    {
	double[] result = new double[3];
	double[] extrap1,extrap2; double r1,r2;
	extrap1 = extrapolations.get_HAD_Extrapolation1(); r1 = Math.sqrt(extrap1[0]*extrap1[0]+extrap1[1]*extrap1[1]);
	extrap2 = extrapolations.get_HAD_Extrapolation2(); r2 = Math.sqrt(extrap2[0]*extrap2[0]+extrap2[1]*extrap2[1]);

	if (r1<r && r<r2) {
	    double fraction = (r - r1) / (r2 - r1);
	    for (int i=0;i<3;i++) {
		result[i] = extrap1[i] + fraction * (extrap2[i] - extrap1[i]);
	    }
	    if (firstHAD_Barrel_Extrapolation) {
		System.out.println("   For cluster at r,z = "+r+","+z);
		// System.out.println("    and search track extrapolations x,y(,r),z of: \n"+
		// 		   " \t\t"+extrap1[0]+","+extrap1[1]+"("+r1+"),"+extrap1[2]+"\n"+
		// 		   " \t\t"+extrap2[0]+","+extrap2[1]+"("+r2+"),"+extrap2[2]);
		System.out.println("    best track extrapolation is "+result[0]+","+result[1]+"("+
				   Math.sqrt(result[0]*result[0]+result[1]*result[1])+"),"+result[2]);
		firstHAD_Barrel_Extrapolation = false;
	    }
	}
	else
	if (Math.abs(extrap1[2])<Math.abs(z) && Math.abs(z)<Math.abs(extrap2[2])) {
	    double fraction = (z - extrap1[2]) / (extrap2[2] - extrap1[2]);
	    for (int i=0;i<3;i++) {
		result[i] = extrap1[i] + fraction * (extrap2[i] - extrap1[i]);
	    }
	    if (firstHAD_Endcap_Extrapolation) {
		System.out.println("   For cluster at r,z = "+r+","+z);
		// System.out.println("    and search track extrapolations x,y(,r),z of: \n"+
		// 		   " \t\t"+extrap1[0]+","+extrap1[1]+"("+r1+"),"+extrap1[2]+"\n"+
		// 		   " \t\t"+extrap2[0]+","+extrap2[1]+"("+r2+"),"+extrap2[2]);
		System.out.println("    best track extrapolation is "+result[0]+","+result[1]+"("+
				   Math.sqrt(result[0]*result[0]+result[1]*result[1])+"),"+result[2]);
		firstHAD_Endcap_Extrapolation = false;
	    }
	}
	else {
	    if (firstHAD_Extrapolation_error) {
		System.out.println(" *** Error *** For cluster at r,z = "+r+","+z);
		System.out.println("  Track extrapolations x,y(,r),z are: \n"+
				   " \t\t"+extrap1[0]+","+extrap1[1]+"("+r1+"),"+extrap1[2]+"\n"+
				   " \t\t"+extrap2[0]+","+extrap2[1]+"("+r2+"),"+extrap2[2]);
		firstHAD_Extrapolation_error = false;
	    }
	    for (int i=0;i<3;i++) {
		result[i] = extrap1[i];
	    }
	}
	return result;
    }

    public double get_EM_Extrapolation_X() { return extrapolations.get_EM_X(); }
    public double get_EM_Extrapolation_Y() { return extrapolations.get_EM_Y(); }
    public double get_EM_Extrapolation_R() { return extrapolations.get_EM_R(); }
    public double get_EM_Extrapolation_Z() { return extrapolations.get_EM_Z(); }

    public double get_HAD_Extrapolation_X() { return extrapolations.get_HAD_X(); }
    public double get_HAD_Extrapolation_Y() { return extrapolations.get_HAD_Y(); }
    public double get_HAD_Extrapolation_R() { return extrapolations.get_HAD_R(); }
    public double get_HAD_Extrapolation_Z() { return extrapolations.get_HAD_Z(); }

    public void setTrackExtrapolations(double x, double y, double z)
    {
	extrapolations = new TrackExtrapolations(x,y,z);
    }


    // IHep3Vector messages
    public double x()
    {
	return getPX();
    }
    public double y()
    {
	return getPY();
    }
    public double z()
    {
	return getPZ();
    }
    public double[] v()
    {
	return getMomentum();
    }
    public double mag()
    {
	return Math.sqrt(mag2());
    }
    public double mag2()
    {
	return VecOp.dot(this,this);
    }
    public double cosTheta() 
    {
	return z()/mag();
    }


    public static boolean firstEM_Barrel_Extrapolation = true;
    public static boolean firstEM_Endcap_Extrapolation = true;

    public static boolean firstHAD_Barrel_Extrapolation = true;
    public static boolean firstHAD_Endcap_Extrapolation = true;

    public static boolean firstEM_Extrapolation_error = true;
    public static boolean firstHAD_Extrapolation_error = true;

    private static double halfpi = 0.5 * Math.PI;
}
