/* AbstractTracker.java

		Reconstruct tracks

 Created by  Mike Ronan  Nov '98   Pick up LCD events within JAS framework and
                                   pass 3D points to pattern recognition software.
 Modified by
   Nick Sinev    Mar '99
   Mike Ronan    Mar '00    - Add methods for controlling histogram packages.

 */

package hep.lcd.recon.ztracking;

import hep.lcd.event.*;
import hep.lcd.recon.*;
import hep.lcd.util.driver.*;

import hep.analysis.*;
import hep.physics.*;

import java.io.PrintWriter;   // Used for remote debug printing
import java.io.Serializable;
import java.text.*;    // Formatting
import java.util.*;    // Random number generator


/**   Picks up LCD events within JAS framework and passes 3D points to pattern recognition software. */
public abstract class AbstractTracker extends AbstractProcessor
    implements Tracker
{
    protected int MaxLayers = 160;
    protected int NLayers;
    protected int MaxPoints;
    protected int NLevels;
    protected int MinLevel=0;

    protected double x[];
    protected double y[];
    protected double z[];
    protected boolean KFlag[];
    protected int ipoi[] = new int[MaxLayers];
    protected int ilon[] = new int[MaxLayers]; 
    protected TrackerHit trkHit[];

    protected int NVXDLayers;

    protected int NTracks;


    public AbstractTracker(int MaxPoints, Specifications specifications, TrackFinder finder)
    {
	this(MaxPoints,specifications,finder,false);
    }

    public AbstractTracker(int MaxPoints, Specifications specifications, TrackFinder finder, boolean hist)
    {

	if (trace||debug) {
	  System.out.println(ClassName+"  version 1.1.8.7");
	  System.out.println("  MaxPoints = "+MaxPoints);
	  System.out.println("  Specifications: ");
	  System.out.println("  \t Inner,Outer radii = "+specifications.getInnerRadius()+","+specifications.getOuterRadius());
	  System.out.println("  \t Half-length = "+specifications.getHalfLength());
	  System.out.println("  \t B Field     = "+specifications.getBField());
	}

	this.MaxPoints = MaxPoints;
	this.specifications = specifications;
	this.finder = finder;

	x = new double[MaxPoints+1];
	y = new double[MaxPoints+1];
	z = new double[MaxPoints+1];
	KFlag = new boolean[MaxPoints+1];
	trkHit = new TrackerHit[MaxPoints+1];

	df.setMaximumFractionDigits(3);

	if (!hist) setNoHist();
	else setHist();
	defaultFolder = "/Tracking/"+getType();

	NPrintEvents = 5; NDebugEvents = 1;
	summarize = first || firstEvents || batch;
    }


    /** Setup tracker parameters. */
    public void setupParameters()
    {
	if (debug) System.out.println(ClassName+" Setting up Track finder");
	finder.setupFinder(this);

	NLevels = finder.getNLevels();
	if (debug) System.out.println("  NLayers, NLevels = "+NLayers+", "+NLevels+" (MinLevel = "+MinLevel+")");

	setNLayers();
    }

    /** Set number of layers in tracking system. */
    public void setNLayers()
    {
	if (debug) System.out.println(ClassName+"setNLayers: ");
	//if (debug) System.out.println("  specifications = "+specifications+", "+vxdSpecifications);
	NLayers = specifications.getNLayers();
	if (vxd!=null) { NVXDLayers = vxdSpecifications.getNLayers(); NLayers += NVXDLayers; }
	if (debug) {
	  text = " NLayers = "+NLayers;
	  if (vxd!=null) text+=" including "+NVXDLayers+" vertex detector layers.";
	  System.out.println(text);
	}
    }

    /** Set Vertex Detector. */
    public void setVertexDetector(VertexDetector detector)
    {
	if (debug) System.out.println(ClassName+"setVertexDetector: vxd = "+detector);
	if (VXDIncluded) {
	  vxd = detector;
	  finder.setVertexDetector(detector);
	  vxdSpecifications = vxd.getSpecifications();
	  setNLayers();  // Set NLayers to Tracker plus Vertex Detector
	}
	else System.err.println(ClassName+"setVertexDetector: Warning VXD is NOT being included.");
    }
    /** Set track fitter. */
    public void setFitter(TrackFitter fitter)
    { 
	this.fitter = fitter;
	finder.setFitter(fitter);
    }
    /** Returns Track finder */
    public TrackFinder getTrackFinder()
    {
	return finder;
    }
    public TrackFinder getFinder() { return getTrackFinder(); }


    /** Returns pool of x-cordinates of hits. */
    public double[] getX()
    {
	return x;
    }
    /** Returns pool of y-cordinates of hits. */
    public double[] getY()
    {
	return y;
    }
    /** Returns pool of z-cordinates of hits. */
    public double[] getZ()
    {
	return z;
    }
    /** Returns an array of markers of used hits. */
    public boolean[] getFlag()
    {
	return KFlag;
    }
    /** Returns array of starting point index for each layer. */ 
    public int[] getIpoi()
    {
	return ipoi;
    }
    /** Returns number of points in each layer. */
    public int[] getIlon()
    {
	return ilon;
    }
    /** Returns tracker hits. */
    public TrackerHit[] getHits()
    {
	return trkHit;
    }

    /** Get tracker specifications. */
    public Specifications getSpecifications()
    {
	return (Specifications) specifications;
    }

    /** Get type of Tracker. */
    public abstract String getType();


    /** Get number of track found. */
    public int getNTracks()
    {
	return finder.getNTracks();
    }


    /** Get track parameters for the selected track. */
    public double[] getParameters(int itrk)
    {
	return finder.getParameters(itrk);
    }
    /** Get the distance of closest approach (DOCA) of the track to the interaction point. */
    public double get_d0(int itrk)
    {
	return finder.get_d0(itrk);
    }
    /** Get the phi azimuth of the point of closest approach. */
    public double get_phi0(int itrk)
    {
	return finder.get_phi0(itrk);
    }
    /** Get the reconstructed curvature of the track. */
    public double get_curv(int itrk)
    {
	return finder.get_curv(itrk);
    }
    /** Get the distance in z from the point of closest approach to the interaction point. */
    public double get_z0(int itrk)
    {
	return finder.get_z0(itrk);
    }
    /** Get the tangent of the dip angle of the track. */
    public double get_tandip(int itrk)
    {
	return finder.get_tandip(itrk);
    }


    /** Get total number of points on a track. */
    public int getNPoints(int itrk)
    {
	return finder.getNPoints(itrk);
    }
    /** Get number of vertex detector points on a track. */
    public int getNVXDPoints(int itrk)
    {
	return 0;
    }
    /** Get number of tracker points on a track. */
    public int getNTrkPoints(int itrk)
    {
	return getNPoints(itrk);
    }

    /** Get all the points on the track. */
    public int[] getTrackPoints(int itrk)
    {
	return finder.getTrackPoints(itrk);
    }
    /** Get a space point on the track. */
    public double[] getTrackPoint(int itrk, int ipnt)
    {
	//if (debug && itrk==0 && ipnt<5) System.out.println(ClassName+"getTrackPoint: Get point #"+ipnt+" for track #"+itrk);
	int ntrk = finder.getNTracks();
	if (itrk > ntrk) {
	    System.err.println(ClassName+"getTrackPoint: Requested track "+itrk+" does not exist. "+
			       "Only "+ntrk+" tracks available");
	    double point[] ={0.,0.,0.}; 
            return point;
	}
	int npnts = finder.getNPoints(itrk);
	if (ipnt > npnts) {
	    System.err.println(ClassName+"getTrackPoint: Requested point "+itrk+" does not exist ."+
			       "Only "+npnts+" points available");
            double point[] ={0.,0.,0.}; 
            return point;
	}
	int itra[] = finder.getTrackPoints(itrk);
	int ip = itra[ipnt];
	if (ip >= MaxPoints) {
	    System.err.println(ClassName+"getTrackPoint: *** Error in accessing tracker point "+ip+" ***");
	    double point[] = {0.,0.,0.};
	    return point;
	}
	double point[] = {x[ip],y[ip],z[ip]};
	return point;
    }

    /** this ID is unique for any point in tracking system  */
    public int getTrackHitID(int itrk, int ipnt)
    {
	int ntrk = finder.getNTracks();
	if (itrk > ntrk) {
	  System.err.println(ClassName+" Requested track "+itrk+" does not exist. "+
			     "Only "+ntrk+" tracks available.");
	  return -1;  
	}
	int npnts = finder.getNPoints(itrk);
	if (ipnt > npnts) {
	  System.err.println(ClassName+" Requested point "+ipnt+" does not exist. "+
			     "Only "+npnts+" points available."); 
	  return -1;  
	}
	int itra[] = finder.getTrackPoints(itrk);
	int ip = itra[ipnt];
	return 2*ip + 1;
    }
    /** Get the tracker hit for particular point on a track. */
    public TrackerHit getTrackHit(int itrk, int ipnt)
    {
	//if (debug && itrk==0 && ipnt<5) System.out.println(ClassName+"getTrackHit: Get point #"+ipnt+" for track #"+itrk);
	int ntrk = finder.getNTracks();
	if (itrk > ntrk) {
	    System.err.println(ClassName+"getTrackHit: Requested track "+itrk+" does not exist. "+
			       "Only "+ntrk+" tracks available");
            return null;
	}
	int npnts = finder.getNPoints(itrk);
	if (ipnt > npnts) {
	    System.err.println(ClassName+"getTrackHit: Requested point "+ipnt+" does not exist. "+
			       "Only "+npnts+" points available."); 
            return null;
	}
	int itra[] = finder.getTrackPoints(itrk);
	int ip = itra[ipnt];
	return trkHit[ip];
    }


    /** Get Monte Carlo particle parent. */
    public MCParticle getMCParticle(int itrk)
    {
	return finder.getMCParticle(itrk);
    }
    /** Get Monte Carlo particle parent for a particular point on a track. */
    public MCParticle getPointMCParent(int itrk, int ipnt)
    {
	TrackerHit hit = getTrackHit(itrk,ipnt);
	return hit.getMCParticle();
    }
    /** Get fraction of hits from parent MC particle. */
    public double getMCPFraction(int itrk)
    {
	return finder.getMCPFraction(itrk);
    }
    /** Get fraction of Tracker hits from parent MC particle. */
    public double getMCPTrackerFraction(int itrk)
    {
	return finder.getMCPTrackerFraction(itrk);
    }


    /** Returns status of track fitting. */
    public boolean fitSuccess(int it)
    {
	return finder.fitSuccess(it);
    } 
    /** Get overall chi-squared for the selected track. */
    public double getChi2(int it)
    {
	return finder.getChi2(it);
    }
    /** Get number of degrees of freedom for the selected track. */
    public double getNDF(int it)
    {
	return finder.getNDF(it);
    }
    /** Get track error matrix. */
    public double[][] getErrorMatrix(int it)
    {
	return finder.getErrorMatrix(it);
    }

    /** Update fit status after fitting tracks. */
    public void updateFitStatus(int itrk, TrackFitter fitter)
    {
	//System.out.println(ClassName+"updateFitStatus: ");
	if (fitter == null) System.err.println("Null pointer passed as a fitter!!!");
	finder.updateFitStatus(itrk,fitter);
    }
    /** Replaces track parameters with fitter results. */
    public void replaceTrackParameters(int itrk, double[] newpar)
    {
	finder.replaceTrackParameters(itrk,newpar);
    }


    /** Assigns given points to the track. */
    public int assignPoints(int it, double[] x, double[] y, double[] z, boolean[] KFlag,
			    int nLayers, int[] ipoi, int[] ilon,
			    double Ddzax, double Dmasu, int MaxPts,
			    int ipnt[][],
			    double Sang[], double Chis[], double ChisN[])
    {
	int npteff = finder.assignPoints(it,x,y,z,KFlag,nLayers,ipoi,ilon,
					 Ddzax,Dmasu,MaxPts,
					 ipnt,Sang,Chis,ChisN);
	return npteff;
    }
/*
    public int assignPoints(int it,int nLayers,double Ddzax, double Dmasu, int MaxPts,
			    int ipnt[][],double Sang[], double Chis[], double ChisN[])
    {
	int npteff = finder.assignPoints(it,nLayers,
					 Ddzax,Dmasu,MaxPts,
					 ipnt,Sang,Chis,ChisN);
	return npteff;
    }
*/

    protected void init()
    {
	if (printDebug) out = getDebugStream();
	if (Hist) {
	  text = ClassName + " histogramming options: \n";
	  text += " Hstd = "+Hstd+", Hevts = "+Hevts+", Hhits = "+Hhits+", Htrks = "+Htrks+", Hfinder = "+Hfinder;
	  if (print) System.out.println(text); if (printDebug) out.println(text);
	}
    }

    public void process(LCDEvent event)
    {
	nEvt++;
	if (firstEvents) {
	    //text = ClassName+"process: "+getType()+" event reconstruction. \t\t Event #"+nEvt;
	    text = " "+getType()+" event reconstruction. \t\t Event #"+nEvt;
	    System.out.println(text); if (batch) System.err.println(text);
	}

	// Update detector specification if needed
	// specifications.update(event);
	// if (specifications.areNew()) setupParameters();

	TrackerHits trkhits = event.getTrackerHits();

	if (Hist) {
	    HEvents++;
	    HistogramFolder.setDefaultFolder(defaultFolder);
	    if (Hevts) histogram("NHits/evt").fill(trkhits.getNHits());
	}

	int ncl=0;
	int nlayer[] = new int[NLayers];
	for (int i=0;i<NLayers;i++) { nlayer[i]=0; }

	if (first) System.out.println(" Count # of Tracker hits on each layer.");
	for (Enumeration hits = trkhits.getHits();(hits.hasMoreElements() && ncl<MaxPoints);) {
	    TrackerHit hit = (TrackerHit) hits.nextElement();
	    int layer = hit.getLayer();
	    if (Hhits && HEvents<=MaxHitHistEvts) histogram("Hits/layer").fill(layer);
	    nlayer[layer]++;
	    ncl++;

	    if (printPoints && ncl<=10) {
		System.out.println(" Hit #"+ncl+": layer,n = "+layer+" "+nlayer[layer]);
		// double point[] = hit.getPoint();
		// System.out.println("   x,y,z= "+
		// 		   point[0]+", "+point[1]+", "+point[2]);
		// System.out.println("   r,z= "+
		// 		   Math.sqrt(point[0]*point[0]+point[1]*point[1])+", "+point[2]);
	    }
	}
	if (printPoints && ncl>10) System.out.println(" ...");
	if (first) System.out.println(" Total of "+ncl+" Tracker hits.");

	int nPoints=0;
	ipoi = new int[NLayers];
	ilon = new int[NLayers];

	ipoi[0] = ilon[0] = 0;
	for (int layer=1; layer<NLayers; layer++) {
	    ipoi[layer] = ipoi[layer-1]+nlayer[layer-1];
	    ilon[layer] = 0;
	}


	if (SimulateMeasurementError) {
	    double RPhiResolution = specifications.getRPhiResolution();
	    double ZResolution    = specifications.getZResolution();

	    if (first) {
	      System.out.println(" Get Tracker points and simulate measurement errors.");
	      System.out.println("  Measurement resolutions: rphi = "+RPhiResolution+", z = "+ZResolution);
	    }
	    for (Enumeration hits = trkhits.getHits();(hits.hasMoreElements() && nPoints<MaxPoints);) {
		TrackerHit hit = (TrackerHit) hits.nextElement();
		int layer = hit.getLayer();
		double point[] = hit.getPoint();
		int iptr = ipoi[layer]+ilon[layer];
		double r = Math.sqrt(point[0]*point[0] + point[1]*point[1]);
		double phi = Math.atan2(point[1],point[0]);
		double RPhi = r*phi + random.nextGaussian()*RPhiResolution;
		phi = RPhi / r;
		x[iptr] = r * Math.cos(phi);  y[iptr] = r * Math.sin(phi);
		z[iptr] = point[2] + random.nextGaussian()*ZResolution;
		KFlag[iptr] = false;
		trkHit[iptr] = hit;
		ilon[layer]++;
		nPoints++;
	    }

	} else {
	  if (first)
	      System.out.println("  Note: No measurement resolution smearing in Track reconstruction.");

	  if (first) System.out.println(" Get Tracker points");
	  for (Enumeration hits = trkhits.getHits();(hits.hasMoreElements() && nPoints<MaxPoints);) {
	    TrackerHit hit = (TrackerHit) hits.nextElement();
	    int layer = hit.getLayer();
	    double point[] = hit.getPoint();
	    int iptr = ipoi[layer]+ilon[layer];
	    x[iptr] = point[0];
	    y[iptr] = point[1];
	    z[iptr] = point[2];
	    KFlag[iptr] = false;
	    trkHit[iptr] = hit;
	    ilon[layer]++;
	    nPoints++;
	  }
	}

	if (Hhits && HEvents<=MaxHitHistEvts) {
	  for (int i=0;i<nPoints;i++) {
	    if (HEvents<=1) {
		histogram("Trk y vs.x").fill(x[i],y[i]);
		double r = Math.sqrt(x[i]*x[i]+y[i]*y[i]);
		histogram("Trk r vs.z").fill(z[i],r);
	    }
	    histogram("Trk z").fill(z[i]);
	    double phi = Math.atan2(y[i],x[i]);
	    histogram("Trk phi").fill(phi);
	    //if ((0.<phi && phi<1.) && z[i]>0.)
	    //histogram("Trk wedge r vs.z").fill(z[i],r);
	  }
	}

	// Find tracks at different levels
	if (first) System.out.println(" Find tracks at "+NLevels+" levels.");
	for (int level=MinLevel; level<NLevels; level++) {
	  if (first) System.out.println(" Find tracks at level "+level);

	  finder.findTracks(level,nPoints,NLayers);
	  // finder.findTracks(level,nPoints,x,y,z,KFlag,specifications,NLayers,ipoi,ilon);

	  if (first)
	    System.out.println("  At level "+level+", number of tracks found = "+finder.getNTracks());
	}

	// Get number of tracks found
	NTracks = getNTracks();

	/** Provide standard histograms. */
	if (Hist) {
	  if (Hstd||Htrks) {
	    histogram("NTracks").fill(NTracks);
	    if (HEvents<MaxTrkHistEvts) {
	      for (int i=0;i<NTracks;i++) {
		//histogram("NPoints/trk").fill(getNPoints(i));
		histogram("# Trk points/trk").fill(getNTrkPoints(i));
		if (Htrks) {
	          HistogramFolder.setDefaultFolder("Tracks");
		  histogram("d0").fill(finder.get_d0(i));
		  histogram("phi0").fill(finder.get_phi0(i));
		  histogram("curv").fill(finder.get_curv(i));
		  histogram("tandip").fill(finder.get_tandip(i));
		}
		histogram("z0").fill(finder.get_z0(i));
		if (Htrks)
		    HistogramFolder.setDefaultFolder("..");
	      }
	    }
	  }
	  if (Hist) HistogramFolder.setDefaultFolder("/");
	}

	if (firstEvents && summarize) {
	    text = "";
	    if (first) text = getType()+" reconstruction summary: \n";
	    text += "  Number of tracks found = "+NTracks;
	    System.out.println(text);  if (printDebug) out.println(text);
	    if (batch) System.err.println(text);
	}

	first = false;
	if (nEvt>=NPrintEvents) firstEvents = printPoints = false;

    } // End process


    /** Set Monte Carlo particle parent. */
    public void setMCParticle(int itrk, MCParticle mcp)
    {
	finder.setMCParticle(itrk,mcp);
    }
    /** Put fraction of hits from parent MC particle. */
    public void putMCPFraction(int itrk, double fraction)
    {
	finder.putMCPFraction(itrk,fraction);
    }
    /** Put fraction of Tracker hits from parent MC particle. */
    public void putMCPTrackerFraction(int itrk, double fraction)
    {
	finder.putMCPTrackerFraction(itrk,fraction);
    }

    public void setRPhiResolution(double resolution)
    {
	specifications.setRPhiResolution(resolution);
    }
    public void setZResolution(double resolution)
    {
	specifications.setZResolution(resolution);
    } 

    /** Sets tracker histogram options */
    protected void setHist()  { Hist = ((Hstd||Hhits||Htrks||Hevts) || Hfinder); }

    /** Turns on Tracker production histograms. */
    public void setHprod() { Hstd = true; setHevts(); }
    /** Turns on Tracker test histograms. */
    public void setHtest() { Hstd = true; setHevts(); setHtrks(); setHhits(); }
    /** Turns on Tracker development histograms. */
    public void setHdev() { setHprod(); }

    /** Turns on Tracker track histograms: track parameters (d0,phi0,...) */
    protected void setHtrks() { Htrks = true; setHist(); }
    /** Turns on Tracker hit histograms: 2D y vs.x, r vs.z plots */
    protected void setHhits() { Hhits = true; setHist(); }
    /** Turns on Tracker event histograms: NHits/evt, NHits/layer, etc. */
    protected void setHevts() { Hevts = true; setHist(); }

    /** Turns off all histogram packages */
    public void setNoHist() { Hist = Hstd = Hhits = Htrks = Hevts = Hfinder = false; }


    String ClassName = "AbstractTracker::"; protected String text;
    protected int nEvt = 0;
    protected boolean summarize;
    protected boolean batch;
    protected int NPrintEvents,NDebugEvents;

    // Inline debugging and analysis
    protected boolean debug = false;
    protected boolean trace = false;  // Set for testing new versions

    // Printing
    protected boolean print = true;
    protected boolean printDebug = true;  protected PrintWriter out = null;

    protected boolean first = true;
    protected boolean firstEvents = true;

    protected boolean printTracks = false;  protected int NTrksPrint = 5;
    protected boolean printPoints = false;
    protected boolean printTrkPnt = false;
    protected boolean printTrkPar = false;
    protected boolean printRecTrk = false;

    protected DecimalFormat df = new DecimalFormat();

    // Histogramming
    protected String defaultFolder = "/Tracking";
    protected int HEvents = 0, MaxHistEvts = 1000, MaxHitHistEvts = 10, MaxTrkHistEvts = 100;
    protected boolean Hist;
    protected boolean Hstd  = false;  // Set for tuning of tracking package
    protected boolean Hhits = false;  // Set for tuning tracking parameters
    protected boolean Htrks = false;
    protected boolean Hevts = false;
    protected boolean Hfinder = false;


    protected boolean SimulateMeasurementError = false;
    protected Random random = new Random();

    protected Specifications specifications;

    protected boolean VXDIncluded;
    protected Specifications vxdSpecifications;

    protected TrackFinder finder;
    protected TrackFitter fitter;
    protected VertexDetector vxd;
}
