/* AbstractVertexDetector.java

		Find and reconstruct tracks in vertex detector.

 Created by  Mike Ronan  Dec '98   Assign CCD detector points to tracks.
 Modified by:

 */

package hep.lcd.recon.ztracking;

import hep.lcd.event.*;

import java.io.PrintWriter;   // Used for remote debug printing
import java.util.*;    // Random number generator


/**   Assign CCD detector points to tracks. */
public abstract class AbstractVertexDetector extends hep.lcd.util.driver.AbstractProcessor
    implements VertexDetector
{
    protected int NLayers=5;
    protected int MaxPoints = 1500;
    protected int MaxTracks = 200;
    protected int MaxTrkPts = 25;
    protected int NLevels;
    protected int MinLevel=0;

    protected boolean processed = false; 
    protected int NVXDPnts;

    protected double x[];
    protected double y[];
    protected double z[];
    protected boolean KFlag[];
    protected TrackerHit vxdHit[];

    protected int[][] ipnt;              // Points on each track


    public AbstractVertexDetector(Tracker tracker, boolean hist)
    {
	if (trace||debug) System.out.println(ClassName+"  version 1.1.8");

	this.tracker = tracker;

	if (!hist) setNoHist();
	else setHist();
	defaultFolder = "/Tracking/"+getType();

	NPrintEvents = 5; NDebugEvents = 1;
    }
    /** Sets Vertex Detector histogram options */
    protected void setHist()  { Hist = ((Hstd||Hhits||Htrks) || (Hpnts||Hassn)); }

    /** Turns on Vertex Detector production histograms. */
    public void setHprod() { Hstd = Hevts = true; setHist(); }
    /** Turns on Vertex Detector test histograms. */
    public void setHtest() { Hstd = Hevts = true; setHist(); }
    /** Turns on Vertex Detector development histograms. */
    public void setHdev()  { setHprod(); setHhits(); }

    /** Turns on Vertex Detector hit histograms:  */
    protected void setHhits() { Hhits = true; HEvents = 0; setHist(); }
    /** Turns on Vertex Detector point histograms: track assignment (xy dev.,z res.,...) */
    public void setHpnts() { Hpnts = true; HEvents = 0; setHist(); }

    /** Turns off all histogram packages */
    public void setNoHist() { Hist = Hstd = Hevts = Hhits = Htrks = Hpnts = Hassn = false; }


    /** Get Vertex Detector specifications. */
    public Specifications getSpecifications() { return specifications; }


    /** Get number of points on a track. */
    public int getNPoints(int itrk)
    {
	if (debug) System.out.println(ClassName+"getNPoints: ");
	return ipnt[itrk][0];
    }

    /** Get a space point on the track. */
    public double[] getTrackPoint(int itrk, int ip)
    {
	//if (debug && itrk==0 && ip<5) System.err.println(ClassName+"getTrackPoint: Get point #"+ip+" for track #"+itrk);
	if (ip<1 || ip>ipnt[itrk][0]) {
	    System.err.println(ClassName+"getTrackPoint: *** point index "+ipnt+" out of range.");
	    double point[] = null;
	    return point;
        } 
	int i = ipnt[itrk][ip];
	if (i >= MaxPoints) {
	    System.err.println(ClassName+"getTrackPoint: *** Error in accessing vertex point "+i+" ***");
	    double point[] = null;
	    return point;
	}
	double point[] = {x[i],y[i],z[i]};
	return point;
    }

    /** Get the vertex detector hit for a point on the track. */
    public hep.lcd.event.TrackerHit getTrackHit(int itrk, int ip)
    { 
	//if (debug && itrk==0 && ip<5) System.err.println(ClassName+"getTrackHit: Get VXD hit for point #"+ip+" on track #"+itrk);
	if(ip > ipnt[itrk][0]) {
	    System.err.println(ClassName+"getTrackHit: *** Error in accessing vertex hit "+ip+
			       " for track "+itrk+" w/ npnts = "+ipnt[itrk][0]);
	    return null;
	}
	int i = ipnt[itrk][ip];
	if (i >= MaxPoints) {
	    System.err.println(ClassName+"getTrackHit: *** Error in accessing vertex hit "+i+" ***");
	    return null;
	}
	return vxdHit[i];
    }

    /** The Hit ID is unique for any point in tracking system  */
    public int getTrackHitID(int itrk, int ip)
    {
	//if (debug) System.err.println(ClassName+"getTrackHitID: Get VXD hit ID for point #"+ip+" on track #"+itrk);
	if (ip > ipnt[itrk][0]) {
	    System.err.println(ClassName+" Error in accessing vertex hit: ip="
			       +ip+" while npnts "+ipnt[itrk][0]);
	    return -1;
	}
	int i = ipnt[itrk][ip];
	return 2*i;
    }


    /** Report on histogram options. */
    protected void init()
    {
	if (Hist) {
	  text = ClassName + " histogramming options: \n";
	  text += " Hstd = "+Hstd+", Hhits = "+Hhits+", Htrks = "+Htrks+", Hassn = "+Hassn;
	  System.out.println(text);
	}
    }


    /** Start Vertex detector event processing. */
    public void startEvent(LCDEvent event)
    {
	nEvt++;
	if (first) { text = ClassName+" Start event in vertex detector \t\t Event #"+nEvt; System.out.println(text); }
	processed = false;
	specifications.update(event);
	NVXDPnts = event.getVXDHits().getNHits();
    }


    /** Finish Vertex detector event processing. */
    public void finishEvent()
    {
	if (first) { text = ClassName+" Finish event in vertex detector \t\t Event #"+nEvt; System.out.println(text); }
	int NTracks = tracker.getNTracks();

	x = tracker.getX();
        y = tracker.getY();
        z = tracker.getZ();
	vxdHit = tracker.getHits();

	ipnt = new int[NTracks][];

	for (int it=0; it<NTracks; it++) {
	    assignHits(it);
	}
	processed = true;

	first = false;
	if (nEvt>=NPrintEvents)
	    firstEvents = printPoints = false;
    }

    /** Assign vertex detector hits from combined tracking list. */
    public void assignHits(int itrk)
    {
	//if (debug && itrk==0) System.out.println(ClassName+"assignHits: ");
	int npnts = tracker.getNVXDPoints(itrk);
	ipnt[itrk] = new int[1+npnts];
	ipnt[itrk][0] = npnts;
	if (npnts==0) return;

	// Copy points from tracker list
	int[] pnt = tracker.getTrackPoints(itrk);
	int n = 0;
	for (int i=1;i<pnt[0];i++) {
	    if (pnt[i]<NVXDPnts) { n++; ipnt[itrk][n] = pnt[i]; }
	}
	if (n!=npnts) System.err.println(ClassName+"assignHits: Error for track #"+itrk);
    }


/*
    public void releaseHits(int it)
    {
	for(int i=0; i<ipnt[it][0]; i++) {
	    int pin = ipnt[it][i+1];
	    KFlag[pin] = false;
	}
	ipnt[it][0] = 0;
    }
*/

    String ClassName = "AbstractVertexDetector::";

    protected String text;
    protected int nEvt = 0;
    protected boolean batch;
    protected int NPrintEvents,NDebugEvents;

    // Inline debugging and analysis
    protected boolean debug = false;
    protected boolean trace = false;  // Set for testing new versions

    // Printing
    protected boolean print = true;
    protected boolean printDebug = true;  protected PrintWriter out = null;

    protected boolean first = true;
    protected boolean firstEvents = true;

    protected boolean printTracks = false;    protected int MaxPrtTrk = 5;
    protected boolean printPoints = false;
    protected boolean printVXDPnt = false;

    // Histogramming
    protected String defaultFolder = "/Tracking/VXD";
    protected int HEvents = 0, MaxHistEvts = 1000, MaxHitHistEvts = 10, MaxTrkHistEvts = 100,
	                       MaxPntHistEvts = 10;
    protected boolean Hist;
    protected boolean Hstd  = false;
    protected boolean Hevts = false;
    protected boolean Hhits = false;
    protected boolean Htrks = false;

    protected boolean Hpnts = false;
    protected boolean Hassn = false;


    protected boolean SimulateMeasurementError = false;
    protected Random random = new Random();

    protected Tracker tracker;
    protected Specifications specifications;
}
