/* AddReconstructedTracks.java

		Add reconstructed tracks to event.

 Created by  Mike Ronan  Dec '98   Add reconstructed tracks to event.
 Modified by:
   Nick Sinev    Mar 99

 */

package hep.lcd.recon.ztracking;

import hep.lcd.event.*;
//import hep.lcd.recon.*;
import hep.lcd.util.driver.*;

import hep.analysis.*;
import hep.physics.*;

import java.io.Serializable;
import java.util.*;
import java.text.*;


/**  Add reconstructed tracks to event. */
final public class AddReconstructedTracks extends AbstractProcessor
{
    final int MaxTrackNmbr = 250;
    private final int maxHashCodes = 500;
    int NFtdTrks = 0;
    int NTracks;

    boolean TrackerOnly = false;  // Normally add vertex detector points
    boolean doFitting = true;
    boolean doRecalCand = false;  // Don't recalculate track parameters while fitting

    //private ReconstructedTrack[] ftdtrks = new ReconstructedTrack[MaxTrackNmbr];
    //private int[] mchscds = new int[maxHashCodes];
    //private MCParticle[] mcpts = new MCParticle[maxHashCodes];


    /** Constructor turns off any inline histograms. */
    public AddReconstructedTracks(Tracker tracker,VertexDetector vertexDetector)
    {
	this(tracker,vertexDetector,false);
    }

    /** Standard constructor: */
    public AddReconstructedTracks(Tracker tracker,VertexDetector vertexDetector,boolean hist)
    {
	System.out.println(ClassName+"  version 1.1.8.5");
	System.out.println(" Add reconstructed tracks to event - ");

	this.tracker = tracker;
	this.vertexDetector = vertexDetector;
        doRecalCand = true;
        if (tracker.getType().equals("Combined")) doRecalCand = false;

	df.setMaximumFractionDigits(4);

	if (!hist) setNoHist();
	else setHist();

	NPrintEvents = 5; NDebugEvents = 1;
    }
    /** Sets tracker histogram options */
    protected void setHist() { Hist = Hist || (Hstd || Hdev || Hhits || Htrks); }

    /** Turns on production histograms. */
    public void setHprod() { Hstd = true; setHist(); }
    /** Turns on test histograms. */
    public void setHtest() { Hstd = Htrks = Hdev = true; setHist(); }
    /** Turns on development histograms. */
    public void setHdev() { setHprod(); Hdev = true; setHtrks(); setHist(); }

    /** Turns on histograms in forming track candidates (e.g. deltaZ, ChiSq, etc.). */
    protected void setHtrks() { Htrks = true; setHist(); }

    /** Turns off all histogram packages */
    public void setNoHist() { Hist = Hstd = Htrks = Hhits = Hdev = false; }


    public void setFitting()
    {
	doFitting = true;
    }
    public void setNoFitting()
    {
	doFitting = false;
    }  
    
    public void useTrackerOnly()
    {
	TrackerOnly = true;
    }
    public void useVertexDetector()
    {
	TrackerOnly = false;
    }


    protected void init()
    {
	if (Hist) {
	  text = ClassName + " histogramming options: \n";
	  text += " Hstd = "+Hstd+", Htrks = "+Htrks+", Hhits = "+Hhits+", Hdev = "+Hdev;
	  System.out.println(text);
	}
    }

    public void process(LCDEvent event)
    {
	nEvt++;
	if (first) System.out.println(ClassName+" Add reconstructed tracks to event. \t\t Event #"+nEvt);

	Specifications specifications = (Specifications) tracker.getSpecifications();

	if (Hist) {
	  HEvents++;
	  HistogramFolder.setDefaultFolder(defaultFolder);
	  if (HEvents>MaxHistEvts) Hstd = false;
	  if (Hstd) histogram("Evt #").fill(nEvt);
	}

	// Get number of tracks found
	NTracks = tracker.getNTracks();

        //if (firstEvents) System.out.println("  Tracker found "+NTracks+" tracks");
	if (Hstd) histogram("NTracks").fill(NTracks);

	final Vector v = new Vector(NTracks);

	// Count number of accepted tracks
	int NAccepted = 0, itn = -1;
	NFtdTrks = 0;

	// For each track ...
	for (int it=0;it<NTracks;it++) {
	// for (int it=18;it<19;it++) {    // Evt #1  Spiraling pi+ -> mu+ track  (r0 = 3.2 cm)
	// for (int it=19;it<20;it++) {    // Evt #1  Spiraling electron (r0 = 11 cm)
	  boolean print = (printTracks && it<NTrksPrint);
	  if (print) System.out.println("\nTrack #"+it+"\n---------");
	  if (it>=NTrksPrint) firstTracks = printPoints = false;  // Used for printing points on first few tracks
         
	  // Ask for the next candidate track
	  TrackCandidate tc = getTrackCandidate(it,TrackerOnly,print);

/*
	  // Recalculate track parameters
	  if (!tracker.fitSuccess(it)) {
	    if (doRecalCand) tc.recalculateTrack(specifications);
	  }
*/

	  itn++;
	  TrackParameters parameters = tc.getTrackParameters();

	  // Calculate overall ChiSq for points on the track
	  //tc.calculateChiSq(specifications);

	  // Extrapolate track to EM and HAD Calorimeters
	  //tc.doExtrapolations(parameters,specifications);
	  TrackExtrapolations extrapolations = tc.getTrackExtrapolations();

	  // Get corresponding MCParticle
	  MCParticle mcp = tc.getMCParticle();
	  double mcpFraction = tc.getMCPFraction();
	  double mcpTrackerFraction = tc.getMCPTrackerFraction();

	  // Set corresponding MCParticle in tracker (used by ReconstructedTrack)
	  tracker.setMCParticle(it,mcp);
	  tracker.putMCPFraction(it,mcpFraction);
	  tracker.putMCPTrackerFraction(it,mcpTrackerFraction);

	  if (print && printTrkCand) {
	    System.out.println("  Recalculated track");
	    System.out.println("   parameters: \n"+
			       "    d0     = "+df.format(parameters.get_d0())  +"\t"+"   phi0 = "+df.format(parameters.get_phi0())+"\n"+
			       "    curv   = "+df.format(parameters.get_curv())+"\t"+"   z0   = "+df.format(parameters.get_z0())+"\n"+
			       "    tandip = "+df.format(parameters.get_tandip()));
	  }

	  // Add new track to list
	  ReconstructedTrack track = new ReconstructedTrack(tracker,it,extrapolations); 
         
/*
	  if (notaDublicate(track)) { 
	    ftdtrks[NFtdTrks] = track;
            NFtdTrks++;
	  }
*/
	  NAccepted++;
	  v.addElement(track);

	  if (print && printRecTrk) {
	      System.out.println("  Reconstructed track");
	      System.out.println("    charge = "+track.getCharge()+"\t"+"   origin = "+
				 df.format(track.getOriginX())+", "+df.format(track.getOriginY())+", "+df.format(track.getOriginZ())+"\n"+
				 "    mom.   = "+df.format(track.getMom())+"\t"+"   px,py,pz = "+
				 df.format(track.getMomentumX())+", "+df.format(track.getMomentumY())+", "+df.format(track.getMomentumZ()));
	    /*
	      if (track.reachesRadius(tracker.getSpecifications().getOuterRadius())) {
		System.out.println("    extrap (r,z) = "+df.format(track.get_EM_Extrapolation_R())+", "+df.format(track.get_EM_Extrapolation_Z()));
	      } else if (track.exitsEnd(tracker.getSpecifications().getHalfLength())) {
		System.out.println("    extrap (x,y,z) = "+df.format(track.get_EM_Extrapolation_X())+", "+
				   df.format(track.get_EM_Extrapolation_Y())+", "+df.format(track.get_EM_Extrapolation_Z()));
	      } else {
		System.out.println("    ends at (x,y,z) = "+df.format(track.get_EM_Extrapolation_X())+", "+
				   df.format(track.get_EM_Extrapolation_Y())+", "+df.format(track.get_EM_Extrapolation_Z()));
	      }
	    */
	  }
	  if (print && it==NTrksPrint && NTracks>NTrksPrint) System.out.println(" ...");

	  // Standard checks
	  if (Hstd && HEvents<MaxTrkHistEvts) {
	      histogram("MCP Fraction").fill(mcpFraction);
	      histogram("MCP Tracker fraction").fill(mcpTrackerFraction);

	    if (Hdev) {
	      if (NTracks == 2) {
	        histogram("2Pr charges").fill(track.getCharge());
		histogram("2Pr zorigin").fill(track.getOriginZ());
		histogram("2Pr Mom.").fill(track.getMom());
		histogram("2Pr Q*Mom.").fill(track.getCharge()*track.getMom());
	      }
	    }
	  }

        }  //end of the loop over all found tracks
	if (Hstd) histogram("NAccepted").fill(NAccepted);

/*
	if (debug) System.out.println("Fitter successfully fitted "+NFtdTrks+" tracks");

	// now put fitted (if requested) or just found into Vector

	final Vector v = new Vector(NFtdTrks);
        for (int i=0; i<NFtdTrks; i++)
	    v.addElement(ftdtrks[i]); 
*/
	// Add track list to event header
  	event.put(event.TrackList,new ReconstructedTrackList(v));     

	if (first) {
	  System.out.println(" Tracking summary: ");
	  System.out.println("  Number of tracks added to event = "+NTracks+"\n");
	}

	first = printTracks = printPoints = false;
        if (nEvt>=NPrintEvents) firstEvents = false;
        if (nEvt>=NDebugEvents) debug = false;

    } // End process

/*
    // see ytracking
    private boolean notaDublicate(ReconstructedTrack track)
    {
     boolean duplicate = false;
     TrackParameters newtp = track.getTrackParameters();
     double newchi2 = track.getChi2();
     double newndf = track.getNDF();
     int newind = track.trindex;
     int newpnts = tracker.getNPoints(newind);
     for(int i=0; i<NFtdTrks; i++)
     {
      newpnts = tracker.getNPoints(newind);
      ReconstructedTrack oldtr = ftdtrks[i];
      TrackParameters oldtp = oldtr.getTrackParameters();
      int oind = oldtr.trindex;
      if((Math.abs(oldtp.get_phi0()-newtp.get_phi0())<0.0001) &&
         (Math.abs(oldtp.get_omega()-newtp.get_omega())<0.00001) &&
         (Math.abs(oldtp.get_s()-newtp.get_s())<0.0001)) duplicate = true;
      if(!duplicate)
      {
       int oldpnts = tracker.getNPoints(oind);
       int ndif = Math.abs(oldpnts-newpnts);
       int ncoi=0;
       int ncut = 0;
       for(int j=1; j<newpnts+1; j++)
       {
       
        int nid = tracker.getTrackHitID(newind,j);
        int minp = j-ndif;
        if(minp < 1) minp = 1;
        int maxp = j+ndif+1;
        if(maxp > oldpnts+1) maxp=oldpnts+1; 
        for(int k=minp; k<maxp; k++)
        {
         if((j>newpnts) || (k>oldpnts)) 
         {
          System.out.println("ERROR TPC: "+j+" "+newpnts+" "+k+" "+oldpnts);
         }
         if(nid == tracker.getTrackHitID(oind,k)) ncoi++;
         if(ncoi > ncut) {duplicate = true; break; } 
        }
        if(ncoi > ncut) {duplicate = true; break; }
       }  
       if(ncoi > ncut) duplicate = true;
      } 
      if(!duplicate)
      {
       newpnts = vertexDetector.getNPoints(newind);
       int oldpnts = vertexDetector.getNPoints(oind);
       int ndif = Math.abs(oldpnts-newpnts);
       int ncoi=0;
       int ncut = 1;
       for(int j=1; j<newpnts+1; j++)
       {
        int nid = vertexDetector.getTrackHitID(newind,j);
        int minp = j-ndif;
        if(minp < 1) minp = 1;
        int maxp = j+ndif+1;
        if(maxp > oldpnts+1) maxp=oldpnts+1; 
        for(int k=minp; k<maxp; k++)
        {
         if((j>5) || (k>5)) 
         {
          System.out.println("ERROR VXD: "+j+" "+newpnts+" "+k+" "+oldpnts);
         }
         if(nid == vertexDetector.getTrackHitID(oind,k)) ncoi++;
         if(ncoi > ncut) {duplicate = true; break; } 
        }
        if(ncoi > ncut) {duplicate = true; break; }
       }  
       if(ncoi > ncut) duplicate = true;
      }      
      if(duplicate)
      {
       boolean newbetter = false;
       double oldchi2 = oldtr.getChi2();
       double oldndf = oldtr.getNDF();
       if((oldndf<1.) || (newndf<1.))
       {
        if(newndf >1.) newbetter = true;
       }
       if((oldndf > 0.5) && (newndf > 0.5))
       {
        double oldchd = oldchi2/oldndf;
        double newchd = newchi2/newndf;
        newbetter = (newchd < oldchd);
       }
       if(newbetter) ftdtrks[i] = track;

       break;
      }         
     }
     return !duplicate;  
    }
*/

    public TrackCandidate getTrackCandidate(int it, boolean TrackerOnly, boolean print)
    {
	TrackCandidate trk = new TrackCandidate(it,getHotList(it,TrackerOnly,print),tracker.getParameters(it));
	//if (print) trk.printPoints(5);
	return trk;
    }
	    

    public TrackHotList getHotList(int it, boolean TrackerOnly, boolean prt)
    {
	int nvtx=0, ntrk=0;
	boolean print = prt || printPoints;
	if (vertexDetector!=null && !TrackerOnly) nvtx = vertexDetector.getNPoints(it);
	if (tracker!=null) ntrk = tracker.getNTrkPoints(it);
	int npts = nvtx + ntrk;
	if (debug) System.out.println(ClassName+"getHotList: npts = "+npts+" (nvtx,ntrk = "+nvtx+","+ntrk+")");

	TrackHotList list = new TrackHotList(npts,nvtx,ntrk);

	int i=0;
	if (vertexDetector!=null && !TrackerOnly) {
	  if (debug && it==0) System.out.println(ClassName+"getHotList: Get Vertex hits on track.");
	  VtxHitOnTrk hot;
	  String type = vertexDetector.getType();
	  for (int j=1;j<=nvtx;j++) {
	    hot = new VtxHitOnTrk(type,vertexDetector.getTrackPoint(it,j),
				  vertexDetector.getTrackHit(it,j).getMCParticle());
	    list.put(i,hot);
	    if (printPoints && (j==1 || (printVXDPnt && j<=NPntsPrint)) && firstTracks) hot.print(i);
	    i++;
	  }
	  if (print) {
	    if (((!printVXDPnt && list.nvtx>1) || (printVXDPnt && list.nvtx>NPntsPrint)) && firstTracks) {
	      System.out.println(" ...");
	      hot = new VtxHitOnTrk(type,vertexDetector.getTrackPoint(it,list.nvtx),
				    vertexDetector.getTrackHit(it,list.nvtx).getMCParticle());
	      hot.print(i-1);
	    }
	    if (list.nvtx!=0 && firstTracks) System.out.println("");
	  }
	}

	if (tracker!=null) {
	  if (debug && it==0) System.out.println(ClassName+"getHotList: Get Tracker hits on track.");
	  TrkHitOnTrk hot;
	  String type = tracker.getType();
	  for (int j=1;j<=ntrk;j++) {
	    hot = new TrkHitOnTrk(type,tracker.getTrackPoint(it,j),
				  tracker.getTrackHit(it,j).getMCParticle());
	    list.put(i,hot);
	    if (print && (j==1 || (printTrkPnt && j<=NPntsPrint)) && firstTracks) hot.print(i);
	    i++;
	  }
	  if (print) {
	    if (((!printTrkPnt && list.ntrk>1) || (printTrkPnt && list.ntrk>NPntsPrint)) && firstTracks) {
	      System.out.println(" ...");
	      hot = new TrkHitOnTrk(type,tracker.getTrackPoint(it,list.ntrk),
				    tracker.getTrackHit(it,list.ntrk).getMCParticle());
	      hot.print(i-1);
	    }
	    if (list.ntrk!=0 && firstTracks) System.out.println("");
	  }
	}
	return list;
    }

    public double[][] getPoints(int it, boolean print)
    {
	int nvtx=0, ntrk=0;
	if (vertexDetector!=null) nvtx = vertexDetector.getNPoints(it);
	if (tracker!=null) ntrk = tracker.getNPoints(it);
	int npts = nvtx + ntrk;

	double[][] points = new double[1+npts][];

	points[0] = new double[3];
	points[0][0] = (double) npts;
	points[0][1] = (double) nvtx;
	points[0][2] = (double) ntrk;

	int i=0;
	if (vertexDetector!=null) {
	  for (int j=1;j<=nvtx;j++) {
	    i++;
	    points[i] = vertexDetector.getTrackPoint(it,j);
	    if (print && (j==1 || (printVXDPnt && j<=5)))
		  System.out.println("  VXD point "+j+": x,y,z = "+
				     points[i][0]+", "+points[i][1]+", "+points[i][2]);
	       // System.out.println("  VXD point "+j+": r,z = "+
	       //                    Math.sqrt(points[i][0]*points[i][0]+points[i][1]*points[i][1])+", "+points[i][2]);
	  }
	  if (print && printVXDPnt) {
	    if (i>5) {
		System.out.println("  ..."+"\n"+
				   "  VXD point "+i+": x,y,z = "+
				   points[i][0]+", "+points[i][1]+", "+points[i][2]);
		               //  "  VXD point "+i+": r,z = "+
			       //  Math.sqrt(points[i][0]*points[i][0]+points[i][1]*points[i][1])+", "+points[i][2]);
	    }
	    System.out.println("");
	  }
	}

	if (tracker!=null) {
	  for (int j=1;j<=ntrk;j++) {
	    i++;
	    points[i] = tracker.getTrackPoint(it,j);
	    if (print && (j==1 || (printTrkPnt && j<=5)))
		  System.out.println("  Trk point "+j+": x,y,z = "+
				     points[i][0]+", "+points[i][1]+", "+points[i][2]);
	       // System.out.println("  Trk point "+j+": r,z = "+
	       //                    Math.sqrt(points[i][0]*points[i][0]+points[i][1]*points[i][1])+", "+points[i][2]);
	  }
	  if (print && printTrkPnt) {
	    if (i>5) {
		System.out.println("  ..."+"\n"+
				   "  Trk point "+i+": x,y,z = "+
				   points[i][0]+", "+points[i][1]+", "+points[i][2]);
		               //  "  Trk point "+i+": r,z = "+
			       //  Math.sqrt(points[i][0]*points[i][0]+points[i][1]*points[i][1])+", "+points[i][2]);
	    }
	    System.out.println("");
	  }
	}
	return points;
    }


    String ClassName = "AddReconstructedTracks::"; String text;
    int nEvt = 0;
    boolean batch;
    int NPrintEvents,NDebugEvents;

    // Inline debugging and analysis
    boolean debug = false;

    // Printing
    boolean first = true;
    boolean firstEvents = true;

    boolean firstTracks = true;
    boolean firstReject = false;

    // Print points on tracks
    boolean printPoints = false;  int NPntsPrint = 10;
    boolean printVXDPnt = true;
    boolean printTrkPnt = true;

    // Print tracks
    boolean printTracks = false;  int NTrksPrint = 5;
    boolean printTrkCand = false;
    boolean printFitTrk = false;
    boolean printRecTrk = true;

    DecimalFormat df = new DecimalFormat();


    // Histogramming
    String defaultFolder = "/Tracking/Add";
    int HEvents = 0, MaxHistEvts = 1000, MaxHitHistEvts = 1, MaxTrkHistEvts = 100;
    boolean Hist;
    boolean Hstd  = false;
	
    boolean Htrace = false;
    boolean  HtraceTrk = false;
    boolean  HtraceHit = false;
    boolean Hdev  = false;

    boolean Hhits = false;
    boolean Htrks = false;


    Tracker tracker;
    VertexDetector vertexDetector;
}

/** Reconstructed track list. */
class ReconstructedTrackList implements TrackList, Serializable
{
    public ReconstructedTrackList(Vector v)
    {
	this.v = v;
    }
    public int getNTracks() { return v.size(); }
    public Enumeration getTracks() { return v.elements(); }
    private Vector v;
}
