/* BaseSpecifications.java

                Specifies Tracker dimensions and performance for any detector.

 Created by  Mike Ronan  Jan '99   Used in passing tracker specifications to TrkCandidate
                                   in recalculating track parameters.
 Modified by
   Nick Sinev    Mar '99    - 
   Mike Ronan    Feb '00    - Simplify object model, set BField when getting VXD specifications.

 */

package hep.lcd.recon.ztracking;

import hep.lcd.event.*;
import hep.lcd.geometry.Detector;
import hep.lcd.geometry.PropertySet;
import hep.lcd.geometry.component.Field_Strength;

import java.io.Serializable;


public abstract class BaseSpecifications
    implements Specifications
{
    protected boolean areNew = true;
    protected String root = null;

    protected Detector detector;
    protected int    DetectorType = 1;         // 1 - large, 2 - small

    protected String Type;

    protected double InnerRadius;
    protected double OuterRadius;
    protected double HalfLength;
    protected int    NLayers;
    protected double LayerRadii[];
    protected double LayerHalfLength[];        
    protected double RPhiResolution;
    protected double ZResolution;
    protected int    NSigmaTracking;

    protected double BField;

    double VxdInnerRadius;
    double VxdOuterRadius;
    double VxdHalfLength;
    double TrkInnerRadius;
    double TrkOuterRadius;
    double TrkHalfLength;
    double EMInnerRadius;
    double EMOuterRadius;
    double EMInnerEndPlane;
    double EMOuterEndPlane;
    double HADInnerRadius;
    double HADOuterRadius;
    double HADInnerEndPlane;
    double HADOuterEndPlane;
    double CoilExtrap;
    double NoFieldExtrap;
    double MuonExtrap;
    double MuonEndExtrap;


    public BaseSpecifications()
    {
	this("Undefined");
    }
    public BaseSpecifications(String type)
    {
	if (trace||debug) System.out.println(ClassName+" Set "+Type+" specifications.");

	this.Type = type;
	this.setSpecs();
    }


    public boolean areNew() { return areNew;}
    public Detector getDetector() {return detector;}
    public int    getDetectorType() { return DetectorType;} 

    public String getType() { return Type; }

    public double getInnerRadius() { return InnerRadius; }
    public double getOuterRadius() { return OuterRadius; }
    public double getHalfLength()  { return HalfLength; }

    public double getLayerRadius(int nl) { return LayerRadii[nl]; }
    public double getLayerZ(int nl) { return LayerHalfLength[nl]; }
    //public double getLayerOuterRadius(int nl) {return LayOutRad[nl]; } 
/*
    public double getHalfLength(double radius)
    {
        int lind = 0; 
        for (int i=0; i<NLayers; i++) {
	    if (radius < LayRadii[i]) { 
		lind = i;
		break;
	    }
        }
        double prad = 0.;
        double nrad = LayRadii[lind];
        double phln = 0.;
        double nhln = LrsHalfLength[lind]; 
        if (lind > 0) { 
	    prad = LayRadii[lind-1];
	    phln = LrsHalfLength[lind-1];
        }
	double halfLength = (radius-prad)/(nrad-prad) * (nhln-phln) + phln;
	return halfLength;
    }
*/
    public double getHalfLength(double radius) { return HalfLength; }

    public int    getNLayers() { return NLayers; }
    public double getRPhiResolution() { return RPhiResolution; }
    public double getZResolution() { return ZResolution; }
    public int    getNSigmaTracking() { return NSigmaTracking; }

    /** Get inner radius of Vertex Detector. */
    public double getVxdInnerRadius() { return VxdInnerRadius; }
    /** Get outer radius of Vertex Detector. */
    public double getVxdOuterRadius() { return VxdOuterRadius; }
    /** Get half-length of Vertex Detector. */
    public double getVxdHalfLength()  { return VxdHalfLength; }

    /** Get inner radius of Tracker. */
    public double getTrkInnerRadius() { return TrkInnerRadius; }
    /** Get outer radius of Tracker. */
    public double getTrkOuterRadius() { return TrkOuterRadius; }
    /** Get half-length of Tracker. */
    public double getTrkHalfLength()  { return TrkHalfLength; }

    public double getBField()  { return BField; }

    public double getCoilExtrap() { return CoilExtrap; }
    public double getNoFieldExtrap() { return NoFieldExtrap; }

    /** Get inner radius of central EM calorimeter. */
    public double get_EM_InnerRadius() { return EMInnerRadius; }
    /** Get outer radius of central EM calorimeter. */
    public double get_EM_OuterRadius() { return EMOuterRadius; }
    /** Get inner end plane of endcap EM calorimeter. */
    public double get_EM_InnerEndPlane() { return EMInnerEndPlane; }
    /** Get outer end plane of endcap EM calorimeter. */
    public double get_EM_OuterEndPlane() { return EMOuterEndPlane; }

    /** Get inner radius of central HAD calorimeter. */
    public double get_HAD_InnerRadius() { return HADInnerRadius; }
    /** Get outer radius of central HAD calorimeter. */
    public double get_HAD_OuterRadius() { return HADOuterRadius; }
    /** Get inner end plane of endcap HAD calorimeter. */
    public double get_HAD_InnerEndPlane() { return HADInnerEndPlane; }
    /** Get outer end plane of endcap HAD calorimeter. */
    public double get_HAD_OuterEndPlane() { return HADOuterEndPlane; }

    public double getMuonExtrap()    { return MuonExtrap; }
    public double getMuonEndExtrap() { return MuonEndExtrap; }


    /** Set detector r-phi resolution. */
    public void setRPhiResolution(double resolution) { RPhiResolution = resolution; }
    /** Set detector z resolution. */
    public void setZResolution(double resolution) { ZResolution = resolution; }

    public String getSystem()
    {
	if (Type=="CCD") {
	    return "VertexDetector";
	}
	else {
	    return "Tracker";
	}
    }

    public void setSpecs()
    {
      if (debug) System.out.println(ClassName+"setSpecs: Type = "+Type);

      PropertySet propertySet;

      detector = Detector.instance();

      if (getSystem()=="VertexDetector") {

        if (debug) System.out.println(" Get Vertex Detector (Vxd) dimensions.");

	propertySet = (PropertySet) detector.get("MultiLayer_VERTX_Barrel");

	if(propertySet != null) {

          VxdInnerRadius = InnerRadius = propertySet.getDouble("inner_radius1");

	  double nob = propertySet.getDouble("num_of_barrels");
	  int inob = (int) nob;

	  VxdOuterRadius = OuterRadius = propertySet.getDouble("outer_radius"+inob);
	  VxdHalfLength  = HalfLength = propertySet.getDouble("z_outer"+inob);

	  NLayers = 0;
	  for(int i =1; i<inob+1; i++) {
            double strad = propertySet.getDouble("inner_radius"+i);
	    double endrad = propertySet.getDouble("outer_radius"+i);
	    double nlpb = propertySet.getDouble("layers"+i);
	    double lthick = (endrad-strad)/nlpb;
	    int inlpb = (int) nlpb;
	    for(int j=0; j<inlpb; j++) {
/*
              if(NLayers > MaxLayers-1) {
                System.out.println("CCDSpecifications::error in number of layers in CCD!");
                NLayers=MaxLayers-1;
              }
*/
	      /// LayRadii[NLayers]=strad+0.5*lthick+j*lthick;
	      NLayers++;
	    }
	  }

	  if (debug) System.out.println(" Get Tracker (Trk) dimensions.");
	  propertySet = (PropertySet) detector.get("MultiLayer_TRACK_Barrel");
	  int NTrkLayers= 0;
	  if (propertySet != null) {
	    TrkInnerRadius = propertySet.getDouble("inner_radius1");
	    nob = propertySet.getDouble("num_of_barrels");
	    inob = (int) nob;
	    for (int i =1; i<inob+1; i++) {
              double nlpb = propertySet.getDouble("layers"+i);
	      int inlpb = (int) nlpb;
	      NTrkLayers = NTrkLayers+ inlpb;
	    }
	    TrkOuterRadius = propertySet.getDouble("outer_radius"+inob);
	    TrkHalfLength  = propertySet.getDouble("z_outer"+inob);
	  }
	  else System.err.println(ClassName+" Error in getting Tracker (Trk) dimensions.");

	  if (debug) System.out.println("  Tracker barrels = "+inob+", layers = "+NTrkLayers); 
	  boolean large = (NTrkLayers > 10);

	  if (large) { DetectorType = 1; BField = 3.0;
	    if (debug) System.out.println("  For Large detector (type "+DetectorType+")  BField = "+BField);
	  } else {
	    DetectorType = 2; BField = 6.0;
	    if (debug) System.out.println("  For Small detector (type "+DetectorType+")  BField = "+BField);
	  }           
	}
      }

      else if (getSystem()=="Tracker") {

        if (debug) System.out.println(" Get Tracker (Trk) dimensions.");

	propertySet = (PropertySet) detector.get("MultiLayer_TRACK_Barrel");

	if (propertySet != null) {
          TrkInnerRadius = InnerRadius = propertySet.getDouble("inner_radius1");

	  double nob = propertySet.getDouble("num_of_barrels");
	  double nl1 = propertySet.getDouble("layers1");
	  int inob = (int) nob;
	  int inl1 = (int) nl1;
	  NLayers = 0;
	  for (int i =1; i<inob+1; i++) {
            double nlpb = propertySet.getDouble("layers"+i);
	    int inlpb = (int) nlpb;
	    NLayers = NLayers+ inlpb;
	  }
	  if (debug) System.out.println("  Tracker barrels = "+inob+", layers = "+NLayers); 
	  boolean large = (NLayers > 10);

	  TrkOuterRadius = OuterRadius = propertySet.getDouble("outer_radius"+inob);
	  TrkHalfLength  = HalfLength  = propertySet.getDouble("z_outer"+inob);

	  if (large) {
            DetectorType = 1;
	    BField = 3.0;
	    if (debug) System.out.println("  For Large detector (type "+DetectorType+")  BField = "+BField);
	  } else {
	    DetectorType = 2;
	    BField = 6.0;
	    if (debug) System.out.println("  For Small detector (type "+DetectorType+")  BField = "+BField);
	  }           

	  if (debug) System.out.println(" Get Vertex Detector (Vxd) dimensions.");
	  propertySet = (PropertySet) detector.get("MultiLayer_VERTX_Barrel");
	  if (propertySet != null) {
	    VxdInnerRadius = propertySet.getDouble("inner_radius1");
	    nob = propertySet.getDouble("num_of_barrels");
	    inob = (int) nob;
	    VxdOuterRadius = propertySet.getDouble("outer_radius"+inob);
	    VxdHalfLength  = propertySet.getDouble("z_outer"+inob);
	  }
	  else System.err.println(ClassName+" Error in getting Vertex Detector (Vxd) dimensions.");
	}
      }

      if (debug) System.out.println(" Get Electromagnetic Calorimeter (EM) dimensions.");
      propertySet = (PropertySet) detector.get("EM_Barrel");
      EMInnerRadius = propertySet.getDouble("inner_radius");
      EMOuterRadius = propertySet.getDouble("outer_radius");

      propertySet = (PropertySet) detector.get("EM_EndCap_North");
      EMInnerEndPlane = propertySet.getDouble("z_inner");
      EMOuterEndPlane = propertySet.getDouble("z_outer");

      if (debug) System.out.println(" Get Hadronic Calorimeter (HAD) dimensions.");
      propertySet = (PropertySet) detector.get("HAD_Barrel");
      HADInnerRadius = propertySet.getDouble("inner_radius");
      HADOuterRadius = propertySet.getDouble("outer_radius");

      propertySet = (PropertySet) detector.get("HAD_EndCap_North");
      HADInnerEndPlane = propertySet.getDouble("z_inner");
      HADOuterEndPlane = propertySet.getDouble("z_outer");

      if (debug) System.out.println(" Get Muon detector (Muon) dimensions.");
      propertySet = (PropertySet) detector.get("MU_Barrel");
      MuonExtrap = propertySet.getDouble("inner_radius")+10.;

      propertySet = (PropertySet) detector.get("MU_EndCap_North");
      MuonEndExtrap = propertySet.getDouble("z_inner")+10.; 

      // To do:
      // This will work after Tony will make handler for Coil section of detector.ini file
      // PropertySet cop = (PropertySet) detector.get("Coil");
      // CoilExtrap = cop.getDouble("r_inner") + 1.;
      // NoFieldExtrap = cop.getDouble("r_outer");
      //   
      // debug = false;
    }

    public void update(LCDEvent e)
    {
	areNew = false;
/*	String newRoot = (String) e.get("DetectorRoot");
	if ((newRoot != null) && !newRoot.equals(root)) {
	    root = newRoot;
	    this.setSpecs();
	    areNew = true;
	}
*/
	Detector dnew = e.getCurrentDetector();
	// System.out.println(" Update:: Detector is now "+dnew);
	if (dnew != detector) {
	    System.out.println(" Update:: Changing from "+detector+" to "+dnew);
	    detector = dnew;
	    this.setSpecs();
	    areNew = true;
	}
    }
/*
    public void findType()
    {
	if (detector == null) detector = Detector.instance();
	// System.out.println(" findType:: detector is now "+detector);
	PropertySet tpr = (PropertySet) detector.get("MultiLayer_TRACK_Barrel");
	if (tpr != null) {
	    double nob = tpr.getDouble("num_of_barrels");
	    double nl1 = tpr.getDouble("layers1");
	    int inob = (int) nob;
	    int inl1 = (int) nl1;
	    NLayers = 0;
	    for (int i =1; i<inob+1; i++) {
		double nlpb = tpr.getDouble("layers"+i);
		int inlpb = (int) nlpb;
		NLayers = NLayers+ inlpb;
	    }
	    if (debug) System.out.println("Specifications: Trker barrels="+inob+" trker layers="+NLayers); 
	    boolean large = (NLayers > 10);
	    if (large) {
		DetectorType = 1;
		RPhiResolution = 0.01;
		ZResolution = 0.14;
	    }
	    else {
		DetectorType = 2;
		RPhiResolution = 0.003;
		ZResolution = 0.003;
	    }
	    if (debug) System.out.println("Specifications: detector type "+DetectorType);
	}

	PropertySet ecbp = (PropertySet) detector.get("EM_Barrel");
	EMCalExtrap = ecbp.getDouble("inner_radius") + 5.;
	if (debug) System.out.println("EMCalExtrap = "+EMCalExtrap);
	PropertySet ecep = (PropertySet) detector.get("EM_EndCap_North");
	if (debug & (ecep == null)) System.out.println("No EM_EndCap section!");
	EMEndExtrap = ecep.getDouble("z_inner") + 5.;
	if (debug) System.out.println("EMEndExtrap = "+EMEndExtrap);
	PropertySet hcbp = (PropertySet) detector.get("HAD_Barrel");
	if (debug & (hcbp == null)) System.out.println("No HAD_Barrel section!");
	HadCalExtrap = hcbp.getDouble("inner_radius") + 10.;
	if (debug) System.out.println("HadCalExtrap = "+HadCalExtrap);
	PropertySet hcep = (PropertySet) detector.get("HAD_EndCap_North");
	if (debug & (hcep == null)) System.out.println("No HAD_EndCap section!");
	HadEndExtrap = hcep.getDouble("z_inner") + 10.;
	if (debug) System.out.println("HadEndExtrap = "+HadEndExtrap);
	PropertySet mubp = (PropertySet) detector.get("MU_Barrel");
	if (debug & (mubp == null)) System.out.println("No MU_Barrel section!");
	MuonExtrap = mubp.getDouble("inner_radius")+10.;
	if (debug) System.out.println("MuonExtrap = "+MuonExtrap);
	PropertySet muep = (PropertySet) detector.get("MU_EndCap_North");
	if (debug & (muep == null)) System.out.println("No MU_EndCap section!");
	MuonEndExtrap = muep.getDouble("z_inner")+10.; 
	if (debug) System.out.println("MuonEndExtrap = "+MuonEndExtrap);

	PropertySet cop = (PropertySet) detector.get("Coil");
	CoilExtrap = cop.getDouble("r_inner") + 1.;
	NoFieldExtrap = cop.getDouble("r_outer");

	Field_Strength field = (Field_Strength) detector.get("Field_Strength");
	BField = field.getField();
    }
*/
    String ClassName = "BaseSpecifications::";

    // Inline debugging and analysis
    protected boolean debug = false;
    protected boolean trace = false;  // Set for testing new versions

}
