/* HelicalSwimmer.java
 *
 * @author Ray Frey - Javatized by Tony Johnson
 *
 */

package hep.lcd.recon.ztracking;

import hep.lcd.event.*;

import hep.physics.*;    // HepVector's

import java.text.*;


/**
 *
 *  Given a particle of initial momentum, position and
 *  charge (p, r0, and iq), calculates its trajectory in a 
 *  uniform magnetic field in the z-direction and determines 
 *  its point of intersection (r_hit) with a cylinder centered at 
 *  the origin with axis parallel to the z-axis and with ends.
 *  Trajectories are parameterized as fn of path length, s.
 *  There are no approximations made.
 *
 *  Units are Tesla, cm, GeV/c
 *
 * @author Mike Ronan - Generalized for track fitting
 * Modified by
 *   Nick Sinev    March 1999    Get path length (needed for mult. scattering calc.)
 *                               also, Angle  handling
 *   Nick Sinev    June 10 1999  Added second intersection point calculations
 */

public class HelicalSwimmer
{

    /**
     *   Create helical swimmer with some default parameters;
     *   (essentually not usable set)
     */
    public HelicalSwimmer()
    {
	this.Bset = false;
	this.Hset = false;
    } 

    /**
     * Create a Helix swimmer for the given trajectory
     * @param parameters - Track parameters
     * @param B      - field strength; uniform, solenoidal, directed along z-axis
     * @param r_cyl  - radius of cylinder
     * @param z_pl   - z-coordinate of planes at end of cylinder (assumes symmetric placement at +/- z )
     */
    public HelicalSwimmer(TrackParameters parameters, double B, double r_cyl, double z_pl)
    {
	this.B = B;
	this.Bset = true;   
	this.r_cyl = r_cyl/100;
	this.z_pl = z_pl/100;
	putParameters(parameters);
	this.Hset = true;
    }
            
    /**
     * Create a Helix swimmer for the given trajectory
     * @param parameters - Track parameters
     * @param r_cyl  - radius of cylinder in cm
     * @param z_pl   - z-coordinate of planes at end of cylinder (assumes symmetric placement at +/- z )
     */
    public HelicalSwimmer(TrackParameters parameters, double r_cyl, double z_pl)
    {
	this.Bset = false;
	this.Hset = true;
	this.r_cyl = r_cyl/100;
	this.z_pl = z_pl/100;
	putParameters(parameters);
    }

    /**
     * Create a Helix swimmer
     * @param B field strength; uniform, solenoidal, directed along z-axis
     * @param r_cyl radius of cylinder
     * @param z_pl z-coordinate of planes at end of cylinder (assumes symmetric placement at +/- z )
     */
    public HelicalSwimmer(double B, double r_cyl, double z_pl)
    {
	this.B = B;
	this.Bset = true;
	this.Hset = false;
	this.r_cyl = r_cyl/100;
	this.z_pl = z_pl/100;
    }

    public void setBField(double BField)
    {
	this.B = BField;
	this.Bset = true;
    }
    /**
     * @return position at point of intersection (x,y,z)
     */
    public double[] getIntersect()
    {
	return r_hit;
    }

    public double[] getIntersect2()
    {
	swim(true);
	return r_hit2;
    }

    /**
     * @return cyl/plane which is first hit
     */
    public int getPlane()
    {
	return i_hit;
    }

    public int getPlane2()
    {
	swim(true);
	return i_hit2;
    }

    public double[] getCenter()
    {
	double[] result = { xc*100, yc*100 };
	return result;
    }
    public double getRadius()
    {
	return R*100;
    }
    /** 
     *  @return track length from initial point to intersection
     */
    public double getLength()
    {
	return path_len;
    }
    
    public double getLength2()
    {
	swim(true);
	return path_len2;
    }

    /*  This crossing angle in XY plane of track with cylinder is needed for calculation of
     * displacement of hit from it's ideal position due to multipple scatering of the particle
     * We calculate deviation of the track point due to multipple scattering in the plane
     * perpendicular to track, and to project it on the cylinder we need to scale it as
     * 1/sin(crossing angle).
     */
    public double getSinXang()
    {
	double rhit =  Math.sqrt(r_hit[0]*r_hit[0] + r_hit[1]*r_hit[1])/100.;

	/*  First we will find cosine square of the angle in XY plane between track and cylinder
	 * surface (or with circle with center in 0,0 and radius equal to distance in XY plane of the 
	 * hit from 0,0 - for the case of intersection with Z plane instead of cylinder).
	 */
	double cosxsq = 0.5*(rhit/R + R/rhit - Rc*Rc/(rhit * R));

	if(cosxsq < 0.0) return 1.;
	if(cosxsq > 1.) return 0.;      
	double sinxa = Math.sqrt(1.-cosxsq);
	return sinxa;
    }

    private void putParticle(MCParticle particle)
    {
	double d0 = 0.;
	double phi0 = Math.atan2(particle.getPY(),particle.getPX());
	double pt = Math.sqrt(particle.getPX()*particle.getPX()+particle.getPY()*particle.getPY());
	double omega = particle.getType().getCharge()*Math.abs(B)*0.3*.01/pt;
	double z0 = 0.;
	double s = particle.getPZ()/pt;

	TrackParameters parameters = new TrackParameters(d0,phi0,omega,z0,s);

	putParameters(parameters);
    }

    private void putParameters(TrackParameters parameters)
    {
	double omega = parameters.get_omega();
	iq = omega>0 ? +1 : -1;
	double d0 = parameters.get_d0();
	phi0 = parameters.get_phi0();
	double tan_lambda = parameters.get_s();
	if (firstSwim) System.out.println("    d0,phi0,tanlam = "+df.format(d0)+","+df.format(phi0)+","+df.format(tan_lambda));

	phic = phi0 + iq * halfpi;
	//double x0 = d0 * Math.cos(phic);  // Wrong! Sign included in both d0 & in phic.
	//double y0 = d0 * Math.sin(phic);
	double x0 = -1. * d0 * Math.sin(phi0);
	double y0 = d0 * Math.cos(phi0);
	double z0 = parameters.get_z0();
	if (firstSwim) System.out.println("    x0,y0,z0 = "+df.format(x0)+","+df.format(y0)+","+df.format(z0));
	Hep3Vector xdoca = new BasicHep3Vector(x0,y0,z0);

	double d0_in = d0/100.;
	r0[0] = x0/100;
	r0[1] = y0/100;
	r0[2] = z0/100;

	// Determine (positive definite) radius of curvature
	if ((Math.abs(omega) > 0.000001) && (Math.abs(omega) < 10.))
	    R  = 0.01/Math.abs(omega);
	    //R  = 0.01/omega;  // Would have to modify swimming code to account for sign.
	if (Math.abs(omega) < 0.000001) R = 10000.;
	if (Math.abs(omega) > 10.) R = 0.001;
	if (firstSwim) System.out.println("    R = "+df.format(R)+"\t\t\t\t Radius of curvature (m)");

	double lambda = Math.atan(tan_lambda);
	cos_lambda = Math.cos(lambda);
	sin_lambda = Math.sin(lambda);
	sin_phi0 = Math.sin(phi0);
	cos_phi0 = Math.cos(phi0);

	double pt = 0.3*R*Math.abs(B);
	//double pt = 0.3...*Math.abs(R)*Math.abs(B);

	if (debug) {
	  double px0 = pt * cos_phi0;
	  double py0 = pt * sin_phi0;
	  double pz0 = pt * tan_lambda;
	  if (firstSwim) System.out.println("    px,py,pz = "+df.format(px0)+","+df.format(py0)+","+df.format(pz0));
	  Hep3Vector pdoca = new BasicHep3Vector(px0,py0,pz0);

	  double xpz = VecOp.cross(xdoca,pdoca).z();
	  if (xpz * d0 > 0.) System.err.println(ClassName+"  DOCA vector is flipped.");
	}

	// Calculate center of circle
	//xc = (d0_in + R) * Math.cos(phic);  // Wrong! Sign included in both d0 & in phic.
	//yc = (d0_in + R) * Math.sin(phic);
	xc = r0[0] + R * Math.cos(phic);
	yc = r0[1] + R * Math.sin(phic);
	//Rc = Math.abs(d0_in + R);
	Rc = Math.sqrt(xc*xc+yc*yc);
	if (firstSwim) System.out.println("    xc,yc = "+df.format(xc)+","+df.format(yc)+"\t\t Center of circle (m)");
	//if (firstSwim) System.out.println("    Rc = "+df.format(Rc));

	this.Hset = true;
    }

    public final static int PROBLEM = 0;
    public final static int PLANE_R = 1;
    public final static int PLANE_PLUSZ = 2;
    public final static int PLANE_MINUSZ = 3;

    /**
     * @param r_cyl  - cylinder to intercept
     */
    public void swim(double r_cyl)
    {
	if (firstSwim) System.out.println(" swim(r="+df.format(r_cyl)+")");
	this.r_cyl = r_cyl/100;
	swam = false;
	swim(false);
    }
    /**
     * @param r_cyl  - cylinder to intercept
     * @param z_pl   - z-coordinate of planes at end of cylinder (assumes symmetric placement at +/- z )
     */
    public void swim(double r_cyl, double z_pl)
    {
	if (firstSwim) System.out.println(" swim(r="+df.format(r_cyl)+",z="+df.format(z_pl)+")");
	this.r_cyl = r_cyl/100;
	this.z_pl = z_pl/100;
	swam = false;
	swim(false);
    }

    /**
     * @param particle - A Monte Carlo particle
     * @param r_cyl  - cylinder to intercept
     * @param z_pl   - z-coordinate of planes at end of cylinder (assumes symmetric placement at +/- z )
     */
    public void swim(MCParticle particle, double r_cyl, double z_pl)
    {
	putParticle(particle);
	this.r_cyl = r_cyl/100;
	this.z_pl = z_pl/100;
	swam = false;
	swim(false);
    }

    /**
     * @param parameters - Standard Track parameters
     */
    public void swim(TrackParameters parameters)
    {
	if (firstSwim) System.out.println("swim(parameters) to r,z = "+df.format(r_cyl*100.)+","+df.format(z_pl*100.));
	putParameters(parameters);
	swam = false;
	swim(false);
    }   
    /**
     * @param parameters - Standard Track parameters
     * @param r_cyl  - cylinder to intercept
     * @param z_pl   - z-coordinate of planes at end of cylinder (assumes symmetric placement at +/- z )
     */
    public void swim(TrackParameters parameters, double r_cyl, double z_pl)
    {
	if (firstSwim) System.out.println("swim(parameters, r,z = "+df.format(r_cyl)+","+df.format(z_pl)+")");
	putParameters(parameters);
	swim(r_cyl,z_pl);
    }

    /**
     * @param p  - 3-momentum (px,py,pz)
     * @param r0  - initial position (x0,y0,z0)
     * @param iq  - charge iq = q/|e| = +/- 1
     */
    public void swim(double[] p, double[] r0_in, int iq)
    {
	if (firstSwim) System.out.println("swim(p,r,iq)");
	double pt2,pt,mom;

	for (int i=0; i<3; i++) { this.p[i] = p[i]; }
	this.iq = iq;

	r0[0] = r0_in[0]/100;
	r0[1] = r0_in[1]/100;
	r0[2] = r0_in[2]/100;
	double d0 = Math.sqrt(r0_in[0]*r0_in[0]+r0_in[1]*r0_in[1]);

	pt2 = p[0]*p[0] + p[1]*p[1];
	pt   = Math.sqrt(pt2);
	mom = Math.sqrt(pt2 + p[2]*p[2]);
	sin_lambda = p[2]/mom;
	cos_lambda = pt/mom;
	cos_phi0   = p[0]/pt;
	sin_phi0   = p[1]/pt;
	phi0       = Math.atan2(p[1],p[0]);
	double tan_lambda = p[2]/pt;
	if (firstSwim) System.out.println("    d0,phi0,tanlam = "+df.format(d0)+","+df.format(phi0)+","+df.format(tan_lambda));

	if (!Bset) {
	    System.out.println("Attempting to use swimmer in momentum mode without setting B field.");
	    R = 10000.;
	}
	R = pt/(0.3*Math.abs(B));
	if (firstSwim) System.out.println("    R = "+df.format(R)+"\t\t\t\t Radius of curvature (m)");

	// Calculate center of circle
	phic = phi0 + iq * halfpi;
	xc = r0[0] + R * Math.cos(phic);
	yc = r0[1] + R * Math.sin(phic);
	Rc = Math.sqrt( xc*xc + yc*yc );
	if (firstSwim) System.out.println("    xc,yc, Rc = "+df.format(xc)+","+df.format(yc)+", "+df.format(Rc));

	Hset = true;

	swam = false;
	swim(false);
    }

    // Primary swim method
    private void swim(boolean swim2)
    {
	//if (firstSwim) System.out.println(" swim("+swim2+")");
	double s = 0.;
	double szhit = 0.;
	double srhit = 0.;
	double sin_lambda_min;
	double darg;

	// to facilitate speed, we will not calculate second intersection point unless
	// requested to do this by swim2 = true           
	if (!swim2 || !swam) {
	    i_hit = PROBLEM;
	    if (!Hset) {
		System.err.println("Attempt to swim without defining Helix parameters!");
		return;
	    }

	    // Calculate useful quantities: lambda is dip angle (w/rt x-y plane;
	    //   phi0 is initial azimuthal angle in x-y plane w/rt x-axis;
	    //   R is radius of circle in x-y plane; 
	    //   Rc**2=xc**2 + yc**2 is distance to center of circle in x-y plane

	    // First, check intercept with planes at z=+/- z_pl
	    //   Eqn of motion z(s)=z0 + s*sin(lambda)
	    // Check +z_pl plane
	    szhit = sbig;
	    sin_lambda_min = z_pl/sbig;
	    if (sin_lambda > sin_lambda_min) {
		i_hit = PLANE_PLUSZ;
		i_hit2 = i_hit;
		szhit = ( z_pl - r0[2] )/sin_lambda;
	    }
	    // otherwise check -z_pl plane  
	    else if (sin_lambda < -sin_lambda_min) {
		i_hit = PLANE_MINUSZ;
		i_hit2 = i_hit;
		szhit = (-z_pl - r0[2] )/sin_lambda;
	    }
	    if (firstSwim) System.out.println("    szhit = "+df.format(szhit)+" (m), for sin(lambda) = "+df.format(sin_lambda));


	    // Now check intercept of x-y plane path with cylinder.
	    // First, check that pt is finite
	    if (R < 0.001) {
		r_hit[0] = 0.;
		r_hit[1] = 0.;
		r_hit[2] = 100*(r0[2] + szhit*sin_lambda);
		System.out.println("Pt is too small !");
		swam2 = false;
		return;
	    }
	    if (firstSwim) System.out.println("   phic= "+df.format(phic)+", Rc= "+df.format(Rc)+
					      ", R= "+df.format(R)+" xc= "+df.format(xc)+" yc= "+df.format(yc)); 
	    if (trace) System.out.println("   Swim R="+df.format(R)+" Rc="+df.format(Rc)+" r="+df.format(r_cyl)); 

	    // Now calculate path length at intersection point
	    darg = Rc/(2.*R) + R/(2.*Rc) - r_cyl*r_cyl/(2.*R*Rc);
	    if (trace||firstSwim) System.out.println("    darg = "+df.format(darg)+" radians.");

	    if (Math.abs(darg) > 1.0) {
		srhit = sbig;
	    } else {
		double diff = halfpi - Math.asin(darg);
		srhit = (R/cos_lambda)*diff;
	    }
	    if (trace||firstSwim) System.out.println("    srhit = "+df.format(srhit)+" (m)");

	    // Determine if cyl or z-plane hit first (smaller path length)
	    if (Math.abs(srhit) <= szhit) {
		s = srhit;
		i_hit = PLANE_R;
		if (trace||firstSwim) System.out.println("    hit = PLANE_R");
	    } else {
		s = szhit;
		if (trace||firstSwim) System.out.println("    hit = PLANE_Z");
	    }
	    if (trace||firstSwim) System.out.println("    s = "+df.format(s)+" m");


	    // Calculate position at this intercept point
	    path_len = s * 100.;
	    double dang = s*cos_lambda/R;
	    if (trace||firstSwim) System.out.println("    dang = "+df.format(dang)+" radians.");
	    if (trace||firstSwim) System.out.println("    phic = "+df.format(phic)+" radians.");
	    darg = phic - iq * (Math.PI-dang);
	    if (trace||firstSwim) System.out.println("    darg = "+df.format(darg)+" radians.");

	    r_hit[0] = 100*(xc + R * Math.cos( darg ));
	    r_hit[1] = 100*(yc + R * Math.sin( darg ));
	    r_hit[2] = 100*(r0[2] + s * sin_lambda);
	    if (firstSwim) System.out.println("   x,y,z = "+df.format(r_hit[0])+","+df.format(r_hit[1])+","+df.format(r_hit[2]));
	    if (firstSwim) System.out.println("   r,phi,z = "+df.format(Math.sqrt(r_hit[0]*r_hit[0]+r_hit[1]*r_hit[1]))+","+
					      df.format(Math.atan2(r_hit[1],r_hit[0]))+","+df.format(r_hit[2]));

	    swam = true;
	    swam2 = false;
	}
	else
	if (swim2) {
	    if (swam2) return;
	    if (R < 0.001) {
		for (int i=0; i<3; i++)
		    r_hit2[i] = r_hit[i];
		path_len2 = path_len;
		swam2 = true;
		return;
	    }

	    double srhit2 = 2. * Math.PI * R - s;   // this is the path length for the
	                                            // second intersection point of helix with cyl.
	    double s2 = szhit;
	    if (srhit2 <= szhit) {
		s2 = srhit2;
		i_hit2 = PLANE_R;
	    }

	    path_len2 = 100. * s2;
	    double dang2 = s2*cos_lambda/R;         // now when we know path length to the second 
	                                            // point we can calculate coordinates the same
                                                    // way as we did it for the first one

	    double darg2 = phic - iq * (Math.PI - dang2); 
	    r_hit2[0] = 100.*(xc + R * Math.cos(darg2));
	    r_hit2[1] = 100.*(xc + R * Math.sin(darg2));
	    r_hit2[2] = 100.*(r0[2] + s2 * sin_lambda);
	    swam2 = true;  
	}

	if (firstSwim) {
	    firstSwim = false;
	    double r = Math.sqrt(r_hit[0]*r_hit[0]+r_hit[1]*r_hit[1]);
	    double phi = Math.atan2(r_hit[1],r_hit[0]);
	    System.out.println("  found intersection: x= "+df.format(r_hit[0])+", y= "+df.format(r_hit[1])+" ("+
			       "r= "+df.format(r)+",phi= "+df.format(phi)+"), z= "+df.format(r_hit[2]));
	} 
    }

    static DecimalFormat df = new DecimalFormat();

    static {
	df.setMaximumFractionDigits(3);
    }

    static String ClassName = "TrackFinder::"; static String text;
    static boolean debug = false;
    static boolean trace = false;

    static boolean firstHelixSwimmer = false;
    static boolean firstSwim = false;

    boolean Bset = false;  // Indicates that magnetic field value set (not always necessary) 
    boolean Hset = false;  // Indicates that helix parameters have been determined

    boolean swam = false;
    boolean swam2 = false;

    double B = 2.0;      // Large detector BField (tesla)
    double r_cyl = 2.0;  //  radius (m)
    double z_pl = 3.0;   //  end plane (m)
    double xc, yc;
    double R = 0.;
    double Rc = 0.;
    double Rq = 0.;
    double sin_lambda = 0.;
    double cos_lambda = 1.;
    double sin_phi0 = 0.;
    double cos_phi0 = 1.;
    double phi0 = 0.;
    double phic = 0.;
    double[] p = new double[3];
    double[] r0 = new double[3];
    double path_len = 0.;
    double path_len2 = 0.;
    int iq = 1;
    static final double  sbig = 99999.;
    static final double  pt_tiny = 0.010;
    double[] r_hit = new double[3];
    double[] r_hit2 = new double[3]; 
    int i_hit;
    int i_hit2;

    static double halfpi = 0.5 * Math.PI;
}
