/* TrackFinder.java

    Defines standard variables and provides default implementation of accessor methods,
  and base implementation of pattern recognition algorithm described below.

    The default TrackFinder is derived from PAT2, which was written by
Henri Videau for PEP-TPC tracks.  The main coding and logic remains from the
original version, though many changes were made over the years by
Gerry Lynch and Orin Dahl.

    The default TrackFinder also inherits from BBPAT2, a derivative made by Gerry Lynch
for SVT track finding in BaBar, and the C++ version BaBarPat2 written by
Natalia Kuznetsova. Some treatments of ambiguities, developed specifically
for BaBar, may still remain in the code and should be reviewed.

 Created by  Mike Ronan  Nov '98   Use 3D points for LCD events within JAS framework to find charged tracks
                                   using BaBarPat2 algorithms and large TPC strategy.
 Modified by:
   Nick Sinev    March '99            
   Nick Sinev    May '99    - Modification of original track finder in attempt to speed it up
   Mike Ronan    May '01    - Add methods for controlling histogram packages.

 */

package hep.lcd.recon.ztracking;

import hep.lcd.event.*;

import hep.analysis.*;
import hep.physics.*;

import Jama.*;

import java.text.*;
import java.util.*;


/**   The default version finds tracks from 3D space points using various triplets of layers. Graded levels
  * of pattern recognition are employed to optimize tracking efficiency.
  */
public abstract class TrackFinder
{
    protected String Finder = "Undefined";

    LCDEvent event = null;

    protected int MaxTracks;
    protected int MaxTrkPts;

    protected final int MaxLayers = 160;
    protected final int MaxPoints = 25000;
    protected final int NAzBins = 50;
    protected final int MaxPntBin = 200;
    protected int MaxLayerToCount = MaxLayers;

    protected double x[]; 
    protected double y[];
    protected double z[];
    protected int ipoi[];
    protected int ilon[];
    protected boolean KFlag[]; 

    protected final int p_d0 = 0;
    protected final int p_phi0 = 1;
    protected final int p_k = 2;
    protected final int p_z0 = 3;
    protected final int p_s = 4;

    protected boolean doFitting = false;
    protected TrackParameters parameters;
    protected HelicalSwimmer swimmer;
    protected boolean useFitter = false;


    public TrackFinder(int MaxTracks, int MaxTrkPts)
    {
	this(MaxTracks,MaxTrkPts,false);
    }
    /** Sets up track finder */
    public TrackFinder(int MaxTracks, int MaxTrkPts, boolean hist)
    {
	if (trace||debug) {
	  String text = ClassName+"  version 1.1.8.5";
	  System.out.println(text); // tracker.out.println(text);
	  System.out.println("  MaxTracks = "+MaxTracks+" MaxTrkPts = "+MaxTrkPts);
	  System.out.println("w/MaxLayers = "+MaxLayers+" MaxPoints = "+MaxPoints);
	}

	this.MaxTracks = MaxTracks;
	this.MaxTrkPts = MaxTrkPts;

	trackPar = new double[MaxTracks][5];         // Track parameters (d0,phi0,omega,z0,tandip)
	itra = new int[MaxTracks][1+MaxTrkPts];      // Points on each track
	path = new double[MaxTracks][1+MaxTrkPts];   // 

	xd = new double[MaxTracks];
	yd = new double[MaxTracks];
	zd = new double[MaxTracks];
	xc = new double[MaxTracks];
	yc = new double[MaxTracks];
	xxc = new double[MaxTracks];
	yyc = new double[MaxTracks];
	si = new double[MaxTracks];
	tz = new double[MaxTracks];
	outp = new boolean[MaxTracks];
	outt = new boolean[MaxTracks];
        fitsuccess = new boolean[MaxTracks];

	ChiSq = new double[MaxTracks];
        NDOF = new double[MaxTracks]; 
        ErrorMatrix = new double[MaxTracks][5][5];

        mcparticle = new MCParticle[MaxTracks];
        mcpFraction = new double[MaxTracks];
        mcpTrackerFraction = new double[MaxTracks];

	parameters = new TrackParameters(0.,0.,0.,0.,0.);

        //df.setMaximumFractionDigits(4);  // microns
        df.setMaximumFractionDigits(3);  // 10's of microns

	if (!hist) setNoHist();
	else setHist();

	NPrintEvents = 5; NDebugEvents = 1;
    }


    /** Setup Finder parameters and initiate callback through the setupParameters method
     * to set pattern recognition strategy. */
    public void setupFinder(Tracker tracker)
    {
	if (debug) System.out.println(ClassName+"::setupFinder: Set Tracker to "+tracker.getType());
	this.tracker = tracker;
        specifications = tracker.getSpecifications();

        //VXDIncluded = false;
        //if (tracker.getType().equals("Combined")) VXDIncluded = true;

	// Set default histogram folder
	//defaultFolder = "/Tracking/"+tracker.getType()+"/"+Finder;
	defaultFolder = Finder;

	// Setup instance parameters which completes in a callback to setup pattern recognition strategy..
	setupParameters();

        swimmer = new HelicalSwimmer(specifications.getBField(),specifications.getOuterRadius(),specifications.getHalfLength());
    }

    /** Get maximum number of tracks allowed. */
    public int getMaxTracks() { return MaxTracks; }
    /** Get number of levels of pattern recognition. */
    public int getNLevels()  { return NLevels; }

    /** Get tracking strategy */
    public TrackingStrategy getTrackingStrategy() { return strategy; }

    public abstract void setupParameters();

    /** Set up pattern recognition strategy. Uses a number of levels (NLevels) and a minimum radius of curvature (minRadius)
     *	at each level to grade the levels of pattern recognition. The number of patterns (NPatterns), the triplets of
     *	layers (triplets) used and the number of points required (nPtsReq) define the track finding strategy.
     */
    public void setupStrategy(int NLevels, double[] minRadius, int NPatterns, int[][] triplets, int[] nPtsReq,
	 		      double rTol, double zTol)
    {
	this.NLevels = NLevels;
	this.MaxLevel = NLevels-1;
	this.minRadius = minRadius;
	this.NPatterns = NPatterns;
	this.triplets = triplets;
	this.nPtsReq  = nPtsReq;
	this.rTol = rTol;
	this.zTol = zTol;

	if (debug) {
	  System.out.println(" "+Finder+" Pattern recognition strategy: ");
	  System.out.println("  Num. Levels  = "+NLevels);
	  System.out.println("  Max. Level = "+MaxLevel);
	  String text=""; for (int i=0;i<NLevels;i++) {
	      text = text+Double.toString(minRadius[i]); if (i<(NLevels-1)) text = text+", "; }
	  System.out.println("  Min. radius of curvature at each level = "+text);
	  System.out.println("  Num. Patterns = "+NPatterns);
	  text=""; for (int i=0;i<NPatterns;i++) { for (int j=0;j<3;j++) {
	      text = text+triplets[i][j]; if (j<2) text = text+","; } if (i<(NPatterns-1)) text = text+"; "; }
	  System.out.println("  Using triplets: "+text);
	  text=""; for (int i=0;i<NPatterns;i++) {
	      text = text+nPtsReq[i]; if (i<NPatterns-1) text = text+","; }
	  System.out.println("  w/ # points req. = "+text);
	  System.out.println("  Vertex tolerances in r,z of "+rTol+", "+zTol);

	  System.out.println(" "+Finder+" Pattern recognition parameters: ");
	  System.out.println("  Max.Tracks,TrkPts = "+MaxTracks+", "+MaxTrkPts);
	}

	if (Hist) {
	    HistogramFolder.setDefaultFolder(defaultFolder);
	    if (Hstd && HEvents<MaxHistEvts) {
		Hchis = new Histogram("HChis/trk");
		HChiSqN = new Histogram("HChiSqN/trk");
	    }
	    HistogramFolder.setDefaultFolder("..");
	}
    };


    /** Returns Tracker which contains Tracker specifications needed in track finding. */
    protected Tracker getTracker()
    {
	return tracker;
    }
    /** Set Vertex Detector. */
    void setVertexDetector(VertexDetector detector)
    {
	if (debug) System.out.println(ClassName+"setVertexDetector: ");
	vxd = detector;
    }
    /** Set track fitter. */
    void setFitter(TrackFitter fitter)
    {
	this.fitter = fitter;
	useFitter = true;
	setupParameters();
    }


    /** Get number of track found. */
    int getNTracks() { return ntra; }

    /** Get track parameters for the selected track. */
    double[] getParameters(int itrk) { return trackPar[itrk]; }
    /** Get the distance of closest approach (DOCA) of the track to the interaction point. */
    double get_d0(int itrk)   { return trackPar[itrk][0]; }
    /** Get the phi azimuth of the point of closest approach. */
    double get_phi0(int itrk) { return trackPar[itrk][1]; }
    /** Get the reconstructed curvature of the track. */
    double get_curv(int itrk) { return trackPar[itrk][2]; }
    /** Get the distance in z from the point of closest approach to the interaction point. */
    double get_z0(int itrk)   { return trackPar[itrk][3]; }
    /** Get the tangent of the dip angle of the track. */
    double get_tandip(int itrk) { return trackPar[itrk][4]; }


    /** Get number of points on a track. */
    int getNPoints(int itrk)  { return itra[itrk][0]; }
    /** Get all the points on the track. */
    int[] getTrackPoints(int itrk)  { return itra[itrk]; }


    /** Get Monte Carlo particle parent. */
    MCParticle getMCParticle(int itrk)
    {
        return mcparticle[itrk];
    } 
    /** Get fraction of hits from parent MC particle. */
    double getMCPFraction(int it)
    {
       return mcpFraction[it];
    }
    /** Get fraction of Tracker hits from parent MC particle. */
    double getMCPTrackerFraction(int it)
    {
       return mcpTrackerFraction[it];
    }


    /** Returns status of track fitting. */
    boolean fitSuccess(int itrk)
    {
        return fitsuccess[itrk];
    }
    /** Get overall chi-squared for the selected track. */
    double getChi2(int itrk)
    {
        return ChiSq[itrk];
    }
    /** Get number of degrees of freedom for the selected track. */
    double getNDF(int itrk)
    {
        return NDOF[itrk];
    }
    /** Get track error matrix. */
    double[][] getErrorMatrix(int itrk)
    {
        return ErrorMatrix[itrk];
    }

    /** Update fit status after fitting tracks. */
    void updateFitStatus(int itrk, TrackFitter fitter)
    {
	boolean fitted = fitter.fitSuccess();
	/// boolean fitted = fitter.getFitSuccess();

	fitsuccess[itrk] = fitted;
	ChiSq[itrk] = fitter.getChi2();
	NDOF[itrk] = fitter.getNDF();
	/// ChiSq[it] = fitter.getChiSq();
	/// NDOF[it] = fitter.getNDOF();
	if (fitted) {
	    replaceTrackParameters(itrk,fitter.getFittedParameters().getParameters());
	    double erm[][] = fitter.getErrorMatrix();
	    for(int i=0; i<5; i++)
		for(int j=0; j<5; j++)
		    ErrorMatrix[itrk][i][j]=erm[i][j];
	}
    }
    /** Replaces track parameters with fitter results. */
    void replaceTrackParameters(int itrk, double[] newpar)
    {
	for (int i=0; i<5; i++)
        trackPar[itrk][i]=newpar[i];
    } 


    /**  Get tracker points from Tracker interface ... */
    public void findTracks(int level, int nPoints, int nLayers)
    {
	x = tracker.getX();
	y = tracker.getY();
	z = tracker.getZ();
	ipoi = tracker.getIpoi();
	ilon = tracker.getIlon();
	KFlag = tracker.getFlag();
	//findTracks(level,nPoints,x,y,z,KFlag,nLayers,ipoi,ilon);
	findTracks(level,nPoints,x,y,z,KFlag,specifications,nLayers,ipoi,ilon);
    }

    /** Given the pattern recognition level and the specified minimum radius of curvature, this method
     * uses various layer triplets to find tracks.
     */
    public void findTracks(int level, int nPoints, double[] x, double[] y, double[] z, boolean[] KFlag,
	//		   int nLayers, int[] ipoi, int[] ilon)
			   Specifications specifications, int nLayers, int[] ipoi, int[] ilon)
    {
	Level = level;

	if (first) System.out.println(ClassName+" Find tracks at level "+Level);

	this.x = x;
	this.y = y;
	this.z = z;
	this.KFlag = KFlag;
	this.ipoi = ipoi;
	this.ilon = ilon;

	InnerRadius = specifications.getInnerRadius();
	OuterRadius = specifications.getOuterRadius();
	HalfLength = specifications.getHalfLength();
	NLayers = specifications.getNLayers();
	RPhiResolution = specifications.getRPhiResolution();
	ZResolution = specifications.getZResolution();
	NSigmaTracking = specifications.getNSigmaTracking();

	dmas = NSigmaTracking*RPhiResolution;
	Dzas = NSigmaTracking*ZResolution;

	if (first) {
	    System.out.println(ClassName+"findTracks: For "+Finder);
	    System.out.println("  Tracker specifications:");
	    System.out.println(" \t Inner,Outer radii = "+InnerRadius+","+OuterRadius);
	    System.out.println(" \t Half-length = "+HalfLength);
	    System.out.println(" \t Num. layers = "+NLayers);
	    System.out.println(" \t  r-phi,z resolution = "+RPhiResolution+","+ZResolution);
	    System.out.println(" \t  # sigma tracking   = "+NSigmaTracking);
	    System.out.println("  using \n"+
			       " \t  dmas for al= "+df.format(dmas));
	    System.out.println(" \t  Dzas       = "+df.format(Dzas));
	    System.out.println(" \t  small      = "+small);
	    System.out.println(" \t  rTol, zTol = "+rTol+", "+zTol);
	    System.out.println("  also \n"+
			       " \t  Dmasu = dmas = "+df.format(dmas));
	    System.out.println(" \t  Ddzax = Dzas = "+df.format(Dzas));
	}
	if (Level >= NLevels) {
	    System.out.println(ClassName+"findTracks: For "+Finder+", max. level is "+(NLevels-1));
	    return;
	}

	int iMin = 0;
	Ibal = nLayers;

	if (Level==0) {
	    //if (first) System.out.println(" At level 0, increment local event counter and clear track count.");
	    nEvt++;
	    HEvents++;  // Needed to control histogramming in public methods
	    ntra  = 0;
	    ntras = 0;
	    if (allEvents) {
		first = true;
		// firstTrack = true;
		firstPoints = true;
		// firstExtra = true;
	    }
	}
	if (Hist) {
	  HistogramFolder.setDefaultFolder(defaultFolder);
	}

	// Set min radius of curvature for a given level, and select a triplet of layers.
	double radiusMin = minRadius[Level];
	if (first) System.out.println(" For min. radius of curvature = "+radiusMin);

	for (Pattern = iMin; Pattern < NPatterns; Pattern++) {
 	    // Get the layers for a given pattern
	    int outerLayer  = triplets[Pattern][0];
	    int middleLayer = triplets[Pattern][1];
	    int innerLayer  = triplets[Pattern][2];
	    int numberPtsRequired = nPtsReq[Pattern];

	    // For this given triplet of layers, find all possible tracks
	    /// selectThreePoints(nPoints, x, y, z, KFlag, nLayers, ipoi, ilon,
	    selectThreePoints(nPoints, nLayers,
			      radiusMin, innerLayer, middleLayer, outerLayer, numberPtsRequired);

	    ntras = ntra; // prepare for next triplet iteration

	} // End the triplet loop

	// System.out.println(" Number of tracks found = "+ntra);
	if (debug && (nEvt>=MaxDebugEvts && Level==(NLevels-1))) debug = false;
	if (first && (!allEvents && !(allLevels && Level<(NLevels-1)))) {
	  if (!debug) {
	    System.out.println(" Finished first pass debugging.");
	    first = false;
     	    firstExtra = false;
	    firstTrack = false;
	    firstPoints = false;
	  }
	}

	// Fill standard histograms
	if (Hstd && HEvents<=MaxTrkHistEvts) {
	  if (Level==(NLevels-1)) {
	    if (Hchis!=null) for (int i=0;i<ntra;i++) Hchis.fill(ChiSq[i]);
	    //if (HchiSq!=null) for (int i=0;i<ntra;i++) HchiSq.fill(ChiSq[i]);
	    //if (HchiSqN!=null) for (int i=0;i<ntra;i++) HchiSqN.fill(ChiSq[i]/itra[i][0]);
	  }
	}
	if (Hist) HistogramFolder.setDefaultFolder("..");

	Level = -1;  // Reset for public method invocations
	return;
    }


    /** Selects three suitable points for a given triplet of layers. */
    /// protected void selectThreePoints(int nPoints, double[] x, double[] y, double[] z, boolean[] KFlag,
	/// 			   int nLayers, int[] ipoi, int[] ilon,
    protected void selectThreePoints(int nPoints,
				     int nLayers,
				     double radiusMin, int innerLayer, int middleLayer, int outerLayer,
				     int numberPtsRequired)
    {
	if (first) System.out.println(ClassName+"selectThreePoints: ");
	if (first) {
	    System.out.println(" For now rPipe/InnerRadius = "+rPipe+"/"+InnerRadius);
	    System.out.println(" For now rOut /OuterRadius = "+rOut+"/"+OuterRadius);
	}

	int howMany=0;
	if (innerLayer != -1) {
	    howMany = ilon[innerLayer]*ilon[middleLayer]*ilon[outerLayer];
	} else {
	    howMany = ilon[middleLayer]*ilon[outerLayer];
	}
	//if (Htrks && HEvents<MaxTrkHistEvts) Histogram.find("howMany").fill(howMany);
	if (first) System.out.println(" Number of combinations in given triplet of layers ("+
				      outerLayer+","+middleLayer+","+innerLayer+") = "+howMany);

	/*  The following test is a protection for noisy data. */
	if (howMany > tooMany) {
	    System.err.println(ClassName+"selectThreePoints: "+Finder+" found too many combinations ("+howMany+") in trial layers "+
			       innerLayer+", "+middleLayer+", "+outerLayer+"\t Evt #"+nEvt);
	    System.err.println("  returning");
	    return;
	}

/*
//set the last layer to consider
                int lastLayerCheck = 0;
                int Ibal = 5;
                while ( ((Ibal>0) && !lastLayerCheck) ) {
                    if (_ilon[Ibal] != 0) {
                        lastLayerCheck = 1;   
                         continue;
                    }
                    Ibal = Ibal - 1;
                }
                if (lastLayerCheck) {
*/

	//  Loop over sets of triplets of layers, only if there are points in ALL layers.
	if (howMany!=0) {

	    // Set the points in the inner layer, including the case when the inner layer is the origin
	    int initialInner=0, finalInner=0;
	    if (innerLayer != -1) {
		initialInner  = ipoi[innerLayer];
		finalInner  = initialInner + ilon[innerLayer] - 1;
	    } else {
		finalInner = initialInner = nPoints;
		x[initialInner] = y[initialInner] = z[initialInner] = 0;
		KFlag[initialInner] = false;
	    }

	    // get first and last points in middle and outer layers
	    int initialOuter = ipoi[outerLayer];
	    int initialMiddle = ipoi[middleLayer];
	    int finalOuter  = initialOuter + ilon[outerLayer] - 1;
	    int finalMiddle = initialMiddle + ilon[middleLayer] - 1;

	    // Loop over points in outer layer:
	    for (int IOuter = initialOuter; IOuter <= finalOuter; IOuter++) {
	      if (firstPoints) System.out.print(" Outer "+(IOuter-initialOuter)+" ");
	      // Make sure this point hasn't been used already
	      if (!KFlag[IOuter]) {
		  if (firstPoints) System.out.println("r,z: = "+
		         df.format(Math.sqrt(x[IOuter]*x[IOuter]+y[IOuter]*y[IOuter]))+", "+df.format(z[IOuter]));

		// Loop over points in inner layer
		for (int IInner = initialInner; IInner <= finalInner; IInner++) {
		  // Make sure this point hasn't been used already
	           if (!KFlag[IInner]) {
		      if (firstPoints) System.out.println("  Inner "+(IInner-initialInner)+" r,z: = "+
		             df.format(Math.sqrt(x[IInner]*x[IInner]+y[IInner]*y[IInner]))+", "+df.format(z[IInner]));

		    // Loop over points in middle layer
		      for (int IMiddle = initialMiddle; IMiddle <= finalMiddle; IMiddle++) {
		      //make sure this point hasn't been used already
	              if (!KFlag[IMiddle]) {

	    // Check that we haven't reach maximum number of tracks
	    if (ntra>=MaxTracks) {
		System.err.println(ClassName+"selectThreePoints: Max. # of tracks reached.");
		return;
	    }

            // OK, got three potentially good points,
            // now get their coordinate differences.
            double P[] = new double[4];
            double Q[] = new double[4];
            double R[] = new double[4];
	    P[1] = x[IInner] - x[IMiddle]; // inner-middle
	    P[2] = x[IMiddle] - x[IOuter]; // middle-outer
	    P[3] = x[IOuter] - x[IInner];  // outer-inner
	    Q[1] = y[IInner] - y[IMiddle]; // inner-middle
	    Q[2] = y[IMiddle] - y[IOuter]; // middle-outer
	    Q[3] = y[IOuter] - y[IInner];  // outer-inner

	    // Get squares of distances between points
	    for (int j = 1; j <=3; j++) R[j] = P[j]*P[j]+Q[j]*Q[j];

	    // Branch out according to which difference is the largest
	    double Rc = 0.5*(R[1]+R[2]+R[3]);
	    int foundBranch = 0;
	    int Im=0, It=0, Jd=0, Jm=0, Jt=0;
	    if ((R[1] > Rc) && (innerLayer!=-1)) {
		foundBranch = 1;       // for this branch, inner(2) and middle(3) points are furthest apart
		Id = IInner;  Jd = 2;
		Im = IOuter;  Jm = 1;
		It = IMiddle; Jt = 3;
	    }
	    if (R[2] > Rc) {
		foundBranch = 2;       // for this branch, outer(1) and middle(3) points are furthest apart
		Id = IMiddle; Jd = 3;
		Im = IInner;  Jm = 2;
		It = IOuter;  Jt = 1;
	    }
	    if ((R[3] > Rc) || ((R[1] > Rc) && (innerLayer==-1)) ) {
		foundBranch = 3;       // here outer(1) and inner(2) points are furthest apart; usual case
		Id = IOuter;  Jd = 1;
		Im = IMiddle; Jm = 3;
		It = IInner;  Jt = 2;
	    }
// C		IF NO DIFFERENCE IS GREATER THAN 1.5 TIMES THE AVERAGE,
// C	GO GET NEW POINTS.  THIS REQUIRES ALL 3 TO BE ON SAME SIDE.
//      Go To 40
	    // Require that the largest difference is greater than 1.5 times the average, i.e. all 3 points are on same side
	    if (foundBranch!=0) {
	      // OK, now do some more tests and do some simple fitting...
	      if (innerLayer == -1) Ddzax = zTol;
	      boolean zCheck = (Math.abs(z[Im]-(z[Id]+z[It])*0.5) < Math.abs((z[Id]-z[It])*0.5)+Ddzax);
	      if (zCheck) {
		// find the area of the triangle
		double area = (P[Jm]*Q[Jt]-P[Jt]*Q[Jm])/2.0;
		boolean areaCheck = Math.abs(area) > small;
		if (areaCheck) {

		  //first of all, find the center of the circle relative to the D point, so
		  //get the coefficients in (x - xd)^2 + (y - yd)^2 + xc(x - xd) + yc(y - yd) = 0
//C	CALCULATE (-TWICE) THE CENTRE OF THE CIRCLE RELATIVE TO D.
//C	COEFICIENTS IN (X-XD)**2+(Y-YD)**2+XC(X-XD)+YC(Y-YD)=0

		  double gamma1 = R[Jt]*Q[Jm] + R[Jm]*Q[Jt];
		  double gamma2 = -(R[Jt]*P[Jm] + R[Jm]*P[Jt]);
		  xc[ntra] = 0.5*gamma1/area;
		  yc[ntra] = 0.5*gamma2/area;
		  double gammam = Math.sqrt(gamma1*gamma1+gamma2*gamma2);
		  double Dd = 0.5*(xc[ntra]*xc[ntra]+yc[ntra]*yc[ntra]);
		  double curv = -4.0*area/gammam;
		  //make sure the radius isn't too small
		  boolean curvCheck = (radiusMin*Math.abs(curv)<1.0);
		  // if (firstTrack) System.out.println("  curv = "+curv);
		  if (curvCheck) {

	    double Psm = Dd - R[Jt];
	    double Pvm = xc[ntra]*Q[Jt] - yc[ntra]*P[Jt];
	    double Pst = Dd - R[Jm];
	    double Pvt = -xc[ntra]*Q[Jm] + yc[ntra]*P[Jm];
	    // Calculate the arc lengths from point D
	    double sm = Math.atan2(Pvm,Psm);
	    double st = Math.atan2(Pvt,Pst);

	    //tan(dipAngle) * radius
	    double dz0;
	    if (innerLayer != -1) {
		// Calculate tan(dip)*radius from points T and D
		tz[ntra] = (z[It] - z[Id])/st;
		// Calculate deviation of point M in z            
		dz0 = tz[ntra]*sm - z[Im] + z[Id];
	    } else {
		// Calculate tan(dip)*radiu from points M and D
		tz[ntra] = (z[Im] - z[Id])/sm;
		// Calculate deviation of point T in z            
		dz0 = tz[ntra]*st - z[It] + z[Id];
	    }
	    // Reject point if this deviation is too large
	    if (innerLayer != -1) Ddzax = Dzas;
	    // if (firstTrack) System.out.println("   dz0 = "+dz0);
	    boolean zDevCheck = (Math.abs(dz0) < Ddzax);
	    if (zDevCheck) {

	      // Limit how much the track can miss z = 0 at r = 0
	      double dzz = tz[ntra]*Math.atan2((yc[ntra]*x[Id]-xc[ntra]*y[Id]),
					 (Dd-x[Id]*x[Id]-y[Id]*y[Id]))+z[Id];
	      // if (firstTrack) System.out.println("   dzz = "+dzz);
	      if (innerLayer == -1 && firstOrigin) {
		  System.out.println("   First origin track: dz0/Ddzax, dzz/zTol = "+
				     dz0+"/"+Ddzax+", "+dzz+"/"+zTol);
		  firstOrigin = false;
	      }
	      //if (Htrks && HEvents<MaxTrkHistEvts) Histogram.find("dzz").fill(dzz);
	      boolean z0Check = (Math.abs(dzz) < zTol);

	      if (z0Check) {
		  if (Hcand && HEvents<MaxCandHistEvts) {
		      // String folder = "Level "+Level+" pattern "+Pattern;
		      // HistogramFolder.setDefaultFolder(folder);
		      Histogram.find("curv pass").fill(curv);
		      Histogram.find("dz0 pass").fill(dz0);
		      Histogram.find("dzz pass").fill(dzz);
		      // HistogramFolder.setDefaultFolder("..");
		  }
		  if (firstTrack)
		      System.out.println("  Track candidate! "+
					 "   Layer points (inner,middle,outer) = "+(IInner-initialInner)+" "+
					 (IMiddle-initialMiddle)+" "+(IOuter-initialOuter));
	      
	    // Get center of circle
	    xxc[ntra] = x[Id] - 0.5*xc[ntra];
	    yyc[ntra] = y[Id] - 0.5*yc[ntra];
	    double di = xxc[ntra]*xxc[ntra] + yyc[ntra]*yyc[ntra];

            // outp is true if track misses pipe
	    outp[ntra] = (di - rPipe - 0.5*Dd)*(di - rPipe - 0.5*Dd) > (2.0*Dd*rPipe);
            // outp is true if track misses inner radius
	    // outp[ntra] = (di - InnerRadius - 0.5*Dd)*(di - InnerRadius - 0.5*Dd) > (2.0*Dd*InnerRadius);
	    if (firstLooper && outp[ntra]) {
		double rrPipe = (di - rPipe - 0.5*Dd)*(di - rPipe - 0.5*Dd)/(2.0*Dd);
		// double rrPipe = (di - InnerRadius - 0.5*Dd)*(di - InnerRadius - 0.5*Dd)/(2.0*Dd);
		System.out.println("   First looper: outp,rrPipe = "+outp+", "+rrPipe);
		firstLooper = false; }
            // outt is true if track leaves detector
	    outt[ntra] = (di - rOut - 0.5*Dd)*(di - rOut - 0.5*Dd) < (2.0*Dd*rOut);
	    // outt[ntra] = (di - OuterRadius - 0.5*Dd)*(di - OuterRadius - 0.5*Dd) < (2.0*Dd*OuterRadius);
	    if (firstLooper && !outt[ntra]) {
		double rrOut = (di - rOut - 0.5*Dd)*(di - rOut - 0.5*Dd)/(2.0*Dd);
		// double rrOut = (di - OuterRadius - 0.5*Dd)*(di - OuterRadius - 0.5*Dd)/(2.0*Dd);
		System.out.println("   First looper: outt,rrOut = "+outt[ntra]+", "+rrOut);
		firstLooper = false;  }

            // Check cross product of point D and center (OD x OC)
	    boolean crsdc = (x[Id]*yyc[ntra] - y[Id]*xxc[ntra])<0.0;

	    boolean crossProdCheck = true;
	    boolean cutPipe = true;
	    boolean outerWall = true;
	    boolean orTest = true;

	    if (innerLayer == -1) {
		if ( (x[IOuter]*x[IMiddle]+y[IOuter]*y[IMiddle])
		       < (x[IMiddle]*x[IMiddle]+y[IMiddle]*y[IMiddle]) ) orTest = false;
	    } else {
		boolean pvotc = (x[It]*yyc[ntra]-y[It]*xxc[ntra])<0.0;
		boolean xorTemp = pvotc^crsdc;
		if (xorTemp) {
		    boolean psodt = (-xxc[ntra]*(xc[ntra] - P[Jm]) - yyc[ntra]*(yc[ntra] - Q[Jm])) > 0;
		    if (!outp[ntra] && psodt) cutPipe = false;   // cuts the pipe!
		    if (outt[ntra] && !psodt) outerWall = false; // arc approaches outer wall and cuts it!
		}
	    }
//	    if (firstTrack && ntra==0)
	    if (firstTrack)
		System.out.println("    AND cutPipe,outerWall = "+cutPipe+", "+outerWall);
	    if (orTest && crossProdCheck && cutPipe && outerWall) {


	    xd[ntra] = x[Id];
	    yd[ntra] = y[Id];
	    zd[ntra] = z[Id];

	    // get the sign of the track
	    si[ntra] = (area>0) ? 1.:-1.;
	    //if (Htrks && HEvents<MaxTrkHistEvts) Histogram.find("sign").fill(si[ntra]);

	    // We have a track - increment track counter.
	    ntra++;

	    // If there are too many tracks, reject the worst ones
	    if (ntra>=MaxTracks) trimTracks(radiusMin);

            //some parameters for makeTrack
	    // double dmas = tkFind->roadFloor(0);
	    // double dmc = tkFind->roadPFactor()/radiusMin;
	    // double Dmasu = dmas + dmc*cFactor*secd1h;
	    double Dmasu = dmas;
	    // Ddzax = Dzas + dzc*cFactor*secd3h;
	    Ddzax = Dzas;

	    /// makeTrack(x,y,z,KFlag,nLayers,ipoi,ilon,numberPtsRequired,
	    makeTrack(nLayers,numberPtsRequired,
		      innerLayer,Ibal,Ddzax,Dmasu,
		      curv,gamma1,gamma2);

	    if (firstTrack) firstTrack = false;


	    } else {
		if (firstFailWC) {
		    System.out.println("  Track failing walls check");
		    firstFailWC = false; }

	    } // walls check

	      } // z0Check
	    } // zDevCheck

		  } // curvCheck
		} // areaCheck
	      } // zCheck
	    } // foundBranch

		      }
		    } // Middle
                  }
		} // Inner
	      }
	    } // Outer
	} // HowMany
    }


    /** Given helix derived from 3 points, this method finds all points on the helix,
     * and if the required minimum number of points are found
     * a track is added to the TrackFinder's internal track list.
     *
     * @param x,y,z                - 
     * @param nLayers              - 
     * @param ipoi                 - 
     * @param ilon                 - 
     * @param numberPtsRequired    - 
     * @param innerLayer           - 
     * @param Ibal                 - 
     * @param Ddzax,Dmasu          - 
     * @param curv                 - 
     * @param gamma1,gamma2        - 
     */
    /* Sets
     *  NPoints,
     *  NDF & ChiSq
     */
    /// protected void makeTrack(double[] x, double[] y, double[] z, boolean[] KFlag,
	/// 		     int nLayers, int[] ipoi, int[] ilon, int numberPtsRequired,
    protected void makeTrack(int nLayers, int numberPtsRequired,
			     int innerLayer, int Ibal,
			     double Ddzax, double Dmasu,
			     double curv, double gamma1, double gamma2)
    {
	int it = ntra-1;
	double Sang[] = new double[1+MaxTrkPts];
	double Chis[] = new double[1+MaxTrkPts];
	double ChisN[] = new double[1+MaxTrkPts];

	if (firstTrack) {
	    System.out.println(ClassName+"makeTrack: For "+Finder);
	    // System.out.println(" it = "+it);
	    // System.out.println(" Ibal = "+Ibal);
	    // System.out.println(" curv = "+curv);
	}

	// Set the number of assigned points on the track to zero
	itra[it][0] = 0;

	// Assign points to the track
	int npteff = assignPoints(it,x,y,z,KFlag,nLayers,ipoi,ilon,
				  Ddzax,Dmasu,MaxTrkPts,
				  itra,Sang,Chis,ChisN);

	if (firstTrack) {
	    System.out.println(" Total number of points on track # "+it+" = "+itra[it][0]);
	    System.out.println(" Num. effective points/required = "+npteff+"/"+numberPtsRequired);
	}

	if (Htrks && HEvents<MaxTrkHistEvts) {
	    Histogram.find("npoints").fill(itra[it][0]);
	    if (Heff) Histogram.find("npteff").fill(npteff);
	 /*
	    String folder = "Level "+Level+" pattern "+Pattern;  // From xtracking AbsTrackFinder
	    HistogramFolder.setDefaultFolder(folder);
	    Histogram.find("npoints").fill(itra[it][0]);
	    if (Heff) Histogram.find("npteff").fill(npteff);
	    HistogramFolder.setDefaultFolder("..");
	 */
	}

	if (npteff < numberPtsRequired) {
	    ntra--;
	}
	else {
            // markUsedHits(KFlag);
 
	    if (Htrks && HEvents<MaxTrkHistEvts) {
		Histogram.find("npoints acc.").fill(itra[it][0]);
		if (Heff) {
		  Histogram.find("npteff  acc.").fill(npteff);
		  if (innerLayer == -1)
		    Histogram.find("origin npteff  acc.").fill(npteff);
		}
	    }

	    // Recalculate orbit for tracks using origin at this point
            if (innerLayer == -1)
	      recalculateTrack(it, x, y, z, Sang, gamma1, gamma2, curv);

	    // Determine track parameters
	    double U = (x[Id]*gamma1 + y[Id]*gamma2)
		/Math.sqrt(gamma1*gamma1+gamma2*gamma2);
	    double term = 2.0*U+curv*(x[Id]*x[Id]+ y[Id]*y[Id]);
	    double d0 = term / (2.0+0.5*curv*term);
	    double arcor = Math.atan2((-xc[it]*yyc[it]+yc[it]*xxc[it]),
				      (-xc[it]*xxc[it]-yc[it]*yyc[it]));

            //  Rc - radius of center of circle C, R - radius of curvature
            double Rc = Math.sqrt(xxc[it]*xxc[it]+yyc[it]*yyc[it]);
	    double R = Math.abs(1/curv);

	    // Distance of closest approach to origin
	    double dR = Math.abs(Rc - R);
	    if (d0<0.) dR = -dR;

	    if (Math.abs(dR - d0)>0.001*Math.abs(dR)) {
		if (NReportClosestApproachErrors>0) {
		    text = ClassName+" dR,d0 = "+df.format(dR)+","+df.format(d0)+
			"\n"+" Setting d0 = dR = "+df.format(dR)+" cm";
		    if (reportClosestApproachErrors) System.err.println(text);
		    else System.out.println(text);
		    NReportClosestApproachErrors--;
		}
		d0 = dR;
	    }
                   
	    // Store track parameters!
	    trackPar[it][p_k] = curv;
	    trackPar[it][p_s] = tz[it]*curv;
	    trackPar[it][p_z0] = z[Id] + tz[it]*arcor;
	    trackPar[it][p_d0] = d0;

	    double V = (x[Id]*gamma2 - y[Id]*gamma1)
		/Math.sqrt(gamma1*gamma1+gamma2*gamma2);
	    double psi = Math.atan2(curv*V,(1.0+curv*U));
	    double av1 = Math.sin(psi)/curv;
	    double cosPsi = Math.cos(psi);
	    double av2;
	    if (cosPsi < 0) av2 = (1.0 - cosPsi)/curv;
	    else av2 = av1*Math.sin(psi)/(1.0 + cosPsi);
	    double beta1 = (curv*(av1*gamma2-av2*gamma1) 
			    + gamma1)/Math.sqrt(gamma1*gamma1+gamma2*gamma2);
	    double beta2 = (-curv*(av1*gamma1+av2*gamma2) 
			    + gamma2)/Math.sqrt(gamma1*gamma1+gamma2*gamma2);
	    trackPar[it][p_phi0] = Math.atan2(-beta1,beta2);

	    // ??
            double SangId=0.;
	    for (int I1=1; I1<=itra[it][0]; I1++) 
	      if (itra[it][I1] == Id) SangId = Sang[I1];
	    for (int I2=1; I2<=itra[it][0]; I2++) {
	      path[it][I2-1] = -arcor/curv + 
	        (Sang[I2] - SangId)/Math.abs(curv);
	    }

	    //if (firstTrack) {
		System.out.println("Track points:");
		System.out.println("  Id x,y,z = "+df.format(x[Id])+","+df.format(y[Id])+","+df.format(z[Id])+"\t Inner point");
		System.out.println("    gamma1 = "+df.format(gamma1)+"\t\t\t ???");
		System.out.println("    gamma2 = "+df.format(gamma2)+"\t\t\t ???");
		System.out.println("    U      = "+df.format(U)+"\t\t\t ???");
		System.out.println("    curv   = "+df.format(trackPar[it][p_k])+"\t\t\t Curvature (1/cm)");
		System.out.println("Track parameters:");
		System.out.println("    d0    = "+df.format(d0)+"\t\t\t Distance of closest approach (xy)");
		if (d0<0.) System.out.println("      Negative value of d0 implies ...");
		System.out.println("    arcor = "+df.format(arcor)+"\t\t\t ??");
		System.out.println("    xc,yc = "+df.format(xc[it])+","+df.format(yc[it])+"\t\t ??");
		System.out.println("    xxc,yyc = "+df.format(xxc[it])+","+df.format(yyc[it])+"\t\t Center of circle (cm)");
		System.out.println("    tz = "+df.format(tz[it])+"\t\t\t\t ??");
		System.out.println("    1/curv = "+df.format(1./trackPar[it][p_k])+"\t\t\t Radius of curvature (cm)");
		System.out.println("  trackPar (d0,phi0,curv,z0,tanLambda): "+df.format(trackPar[it][p_d0])+", "+
				   df.format(trackPar[it][p_phi0])+", "+df.format(trackPar[it][p_k])+", "+
				   df.format(trackPar[it][p_z0])+", "+df.format(trackPar[it][p_s]));
	    //}

            parameters.put_d0(d0);
            parameters.put_omega(curv);
            parameters.put_phi0(trackPar[it][p_phi0]);
            parameters.put_z0(trackPar[it][p_z0]);
            parameters.put_s(trackPar[it][p_s]);
	/**/
	    if (firstTrack) {
	      double xx,yy,zz,rr,resid;
	      xx = x[itra[it][0]]; yy = y[itra[it][0]]; zz = z[itra[it][0]];
	      rr = Math.sqrt(xx*xx+yy*yy);
	      swimmer.swim(parameters,rr,HalfLength);
	      System.out.println(" Track swam with newly created parameters to r = "+df.format(rr)+".");
	/**/
	      if (debug) {
		// Check inner points
	        double[] pp;
		for (int i=1;i<=Math.min(itra[it][0],5);i++) {
		  int ip = itra[it][i];
		  rr = Math.sqrt(x[ip]*x[ip]+y[ip]*y[ip]);
		  swimmer.swim(rr,HalfLength);
		  pp = swimmer.getIntersect();
		  resid = Math.sqrt((x[ip]-pp[0])*(x[ip]-pp[0])+(y[ip]-pp[1])*(y[ip]-pp[1])+(z[ip]-pp[2])*(z[ip]-pp[2]));
		  //System.out.println("  for point at x,y,z = "+df.format(x[ip])+","+df.format(y[ip])+","+df.format(z[ip])+
			//	     " track extrapolates to "+df.format(pp[0])+","+df.format(pp[1])+","+df.format(pp[2])+
			//	     " (resid. = "+df.format(resid)+")");
		  if (resid>0.1) System.err.println(ClassName+" Large residual ("+df.format(resid)+
						    ") between extrapolated track and point #"+i+" on track.");
		}
		// Check large radius point
		if (itra[it][0]>5) {
		  xx = x[itra[it][0]]; yy = y[itra[it][0]]; zz = z[itra[it][0]];
		  rr = Math.sqrt(xx*xx+yy*yy);
		  swimmer.swim(rr,HalfLength);
		  pp = swimmer.getIntersect();
		  resid = Math.sqrt((xx-pp[0])*(xx-pp[0])+(yy-pp[1])*(yy-pp[1])+(zz-pp[2])*(zz-pp[2]));
		  //System.out.println("  for last point at x,y,z = "+df.format(xx)+","+df.format(yy)+","+df.format(zz)+
			//	     " track extrapolates to "+df.format(pp[0])+","+df.format(pp[1])+","+df.format(pp[2])+
			//	     " (resid. = "+df.format(resid)+")");
		  if (resid>0.1) System.err.println(ClassName+" Large residual ("+df.format(resid)+
						    ") between extrapolated track and last point on track.");
		}
	      }
	/**/
	    }
	/**/

	    // Track is not being fitted at this time
	    fitsuccess[it] = false;

	    // Set number of degrees of freedom (NDF)
	    NDOF[it] = itra[it][0];

	    // Define chi-squared for this track
	    ChiSq[it] = 0;
	    for (int I = 1; I <= itra[it][0]; I++) {
		ChiSq[it] = ChiSq[it] + Chis[I];
	    }
	    if (Htrks && HEvents<MaxTrkHistEvts)
		Histogram.find("ChiSq/trk").fill(ChiSq[it]);
	}
	return;
    }


    /**  Assigns points to given track.
     */
    public int assignPoints(int it, double[] x, double[] y, double[] z, boolean Flag[],
			    int nLayers, int[] ipoi, int[] ilon,
			    double Ddzax, double Dmasu, int MaxPts,
			    int ipnt[][],
			    double Sang[], double Chis[], double ChisN[])
    {
	if (firstTrack || firstExtra || (firstVtxPt && MaxPts<MaxTrkPts))
	    System.out.println(ClassName+"assignPoints: "+Finder+"  nLayers = "+nLayers);

	int npteff = 0;

	// Set the number of points on the side of interest to zero
	int npco = 0;
	double pisi = (2.0*Math.PI)*si[it]; // signed 2 pi

	int prevLayer = -1;

	// Loop over all layers.
	for (int layer = 0; layer < nLayers; layer++) {
	    // Save how many points there were before this layer
	    int nPointsBeforeLayer = ipnt[it][0];

	    for (int i = ipoi[layer]; i < ipoi[layer]+ilon[layer]; i++) {
                // Skip points that have been used already
		if (!Flag[i]) {
		  //if (firstExtra || (firstVtxPt && MaxPts<MaxTrkPts)) 
		  //    System.out.println("  Layer "+layer+" #"+(i-ipoi[layer])+" r,z: = "+
			//		 df.format(Math.sqrt(x[i]*x[i]+y[i]*y[i]))+", "+df.format(z[i]));

		  // Get x and y differences from point D.
		  double xt = x[i] - xd[it];
		  double yt = y[i] - yd[it];
		  // Calculate square of distance from point D.
		  double Rdd = xt*xt + yt*yt;
		  // Calculate approximate deviation from circle.
		  double pc = xt*xc[it] + yt*yc[it];
		  double al = (Rdd+pc)/Math.sqrt(xc[it]*xc[it]+yc[it]*yc[it]);
		  //if (firstExtra) System.out.println("   xy dev. "+df.format(al));
		  //if (Hpnts && HEvents<=MaxPntHistEvts) Histogram.find("xy dev.").fill(Math.abs(al));

		  // Continue only if this deviation isn't too large
		  boolean largeDev = Math.abs(al) > Dmasu;

		  if (!largeDev) {
		    if (Hpnts && HEvents<=MaxPntHistEvts) Histogram.find("xy dev. pass").fill(al);

                    //cross product of CD and CP
		    double ps = xc[it]*yt - yc[it]*xt;
		    //turning angle
		    double Dd = 0.5*(xc[it]*xc[it]+yc[it]*yc[it]);
		    double as = Math.atan2(ps,(Dd+pc));
		    //DC x OD = OC x OD
		    double pss = xd[it]*yc[it] - yd[it]*xc[it];

		    // pow true means pipe on the way
		    boolean pow = (pss*si[it])>0.0;
		    // cote is true if the point is not on the same half circle
		    // as point t, with the origin defined by point d.
		    boolean cote = (ps*si[it])<0.0;
		    // OM x OC
		    boolean crspc = (x[i]*yyc[it] - y[i]*xxc[it])>0.0;
		    boolean crsdc = (xd[it]*yyc[it] - yd[it]*xxc[it])<0.0;
		    // cotec is true if the point is on the same side of OC as D
                    boolean cotec = !(crsdc^crspc);
		    //if (firstExtra) System.out.println("  pow,cote,crspc,crsdc,cotec = "+
			//			       pow+" "+cote+" "+crspc+" "+crsdc+" "+cotec);

// C               IF THE CIRCLE DOESN'T ENCOUNTER ANY BOUNDARY, WE MAKE TWO TURNS
		    // consider all but opposite side hits of tracks 
		    // that hit both walls
                    double as2, s, Zz;
		    //if (firstExtra) System.out.println("  OR !outt,outp,!cotec = "+!outt[it]+" "+outp[it]+" "+!cotec);

                    if ((!outt[it]) || outp[it] || (!cotec)) {  // do points on tracks that miss both walls
		      if (outp[it] && (!outt[it])) {
		        boolean xorTemp3,invXorTemp3; 
                        xorTemp3 = outp[it]^pow;
                        if (cote && !xorTemp3) as = as + pisi;
                        if ((!cote) && (xorTemp3)) as = as - pisi;
                        s=as;  // calculate arc length
                      } else {
		        if(cotec) {  // points on opposite side of track
			  if (cote) as2 = as + pisi;
			  else as2 = as - pisi;
			  s = as2;
		        } else {     // points on same side
		          s = as;
		        }
		      }
                      Zz = z[i] - zd[it] - s*tz[it];
		      //if (Hpnts && HEvents<=MaxPntHistEvts) Histogram.find("z res.").fill(Zz);

		      // Require a small Z residual
                      boolean smallZRes = Math.abs(Zz) < Ddzax;

                      if (smallZRes) {
			if (Hpnts && HEvents<=MaxPntHistEvts) Histogram.find("z resid. pass").fill(Zz);


            // We have a good point - increment point counter - unless there are too many points.
            if (ipnt[it][0]<MaxPts) {
		if (layer != prevLayer)
		    npteff++;
		else {
		    //if (debug) System.out.println(" More than one point on layer "+layer);
		}
		Flag[i] = true;
		ipnt[it][0]++;
		prevLayer = layer;
		if (firstExtra) {
		    System.out.println(" Good extra point #"+(i-ipoi[layer])+" added on layer "+layer);
		    if (!allPoints) firstExtra = false;
		}
		if (MaxPts<MaxTrkPts && firstVtxPt) {
		    System.out.println(" Good vertex point #"+(i-ipoi[layer])+" added on layer "+layer);
		    if (!allPoints) firstVtxPt = false;
		}

		// Store point number and ...
		int ip = ipnt[it][0];
		ipnt[it][ip] = i;
		Sang[ip] = -s*si[it];

		// Get a crude chi-squared for the point
		double Dmas = 0.009;
		double Dzas = 0.020;
		Chis[ip] = (al/Dmas)*(al/Dmas) + (Zz/Dzas)*(Zz/Dzas);
		ChisN[ip] = (al/Dmasu)*(al/Dmasu) + (Zz/Ddzax)*(Zz/Ddzax);
		if (Hpnts && nEvt<=MaxPntHistEvts) Histogram.find("Chis[ip]").fill(Chis[ip]);

		if (!cote) npco++;

	    } else {
		if (firstFailMaxP) {
		    System.out.println("  Number of points on track exceeding maximum of "+MaxPts);
		    firstFailMaxP = false;
		}
	    } // <MaxPts


		      } //end the small Z deviation

		    } else {
			if (firstFailBW) {
			    System.out.println("  Track failing both walls check");
			    firstFailBW = false; }
		    } //end miss both walls

		  }//end the largeDev check

		}//end the Flag check

            }//end the number of points on a given layer loop

	    if (ipnt[it][0] - nPointsBeforeLayer > 1)
	        eliminateAmbiguities(it, x, y, z, ipnt, nPointsBeforeLayer,
				     Chis, ChisN, Sang, Dmasu);

        }//end the layer loop

	return npteff;
    }


    /** Looks for ambiguous points on tracks.
     *   If we found more than one point in a given layer ...
     */
    protected void eliminateAmbiguities(int it, double[] x, double[] y, double[] z,
				     int ipnt[][], int nptsbl,
				     double[] Chis, double[] ChisN, double[] Sang, double Dmasu)
    {
	if (firstAmbigu) {
	    System.out.println(ClassName+"eliminateAmbiguities: For "+Finder);
	    System.out.println("  Num. points on layer = "+(ipnt[it][0]-nptsbl));
	    System.out.println("  Dmasu = "+Dmasu);
	}

	firstAmbigu = false;
	return;   
    }


    /** Recalculates track parameters for tracks that go through the origin */
    protected void recalculateTrack(int it, double[] x, double[] y, double[] z,
				  double[] Sang, double gamma1, double gamma2, double curv)
    {
	int iptt=0, iptd=0, iptm=0;

        //first, get the first and last points
	double sp = -10000;
	double sm = 10000;

	if (firstRecal) {
	    System.out.println(ClassName+"recalculateTrack: For "+Finder);
	}

	for (int i = 1; i <= itra[it][0]; i++) {
	    if (Sang[i] < sm) {
		sm = Sang[i];
		iptt = i;
	    }
	    if (Sang[i] > sp) {
		sp = Sang[i];
		iptd = i;
	    }
	}

        //now get the middle point
	double sb = 10000;
	double smid = 0.5*(sp + sm);
	for (int i = 1; i <= itra[it][0]; i++) {
	    if (i != iptt  && i != iptd) {
		if (Math.abs(smid-Sang[i])  <  sb) {
		    sb = Math.abs(smid-Sang[i]);
		    iptm = i;   
		}
	    }
	}
        
	//now, set the new points
	int It = itra[it][iptt];
	Id = itra[it][iptd];
	int Im = itra[it][iptm];

	xd[it] = x[Id];
	yd[it] = y[Id];
	zd[it] = z[Id];

	//get the P,Q,R terms and then the triangle area
	double Pot = x[It] - x[Id];
	double Qot = y[It] - y[Id];
	double Pom = x[Im] - x[Id];
	double Qom = y[Im] - y[Id];

	double Rot = Pot*Pot + Qot*Qot;
	double Rom = Pom*Pom + Qom*Qom;
	double Aire = -(Pot*Qom - Pom*Qot)/2.0;
     
	//don't let the area be zero
	if (Aire == 0.) Aire = 1.0E-06;
	//calculate the center of circle relative to point Id
	xc[it] = -0.5*(Rom*Qot - Rot*Qom)/Aire;
	yc[it] = -0.5*(Rot*Pom - Rom*Pot)/Aire;
	//get the gamma vector that is used for the azimuth calculation
	gamma1 = 2.0*xc[it]*Aire;
	gamma2 = 2.0*yc[it]*Aire;
	double gammam = Math.sqrt(gamma1*gamma1 + gamma2*gamma2);
	//now the absolute center of the circle
	xxc[it] = x[Id] - 0.5*xc[it];
	yyc[it] = y[Id] - 0.5*yc[it];
	double Dd = 0.5*(xc[it]*xc[it] + yc[it]*yc[it]);
	//the curvature
	curv = -4.0 * Aire/gammam;
	//calculate the quantities needed for tz
	double Psd = Dd -x[Id]*xc[it] - y[Id]*yc[it];
	double Pst = xxc[it]*(xxc[it] - x[It]) + yyc[it]*(yyc[it] - y[It]);
	double Pvd = xc[it]*y[Id] - yc[it]*x[Id];
	double Pvt = x[It]*yyc[it] - y[It]*xxc[it];
	double Asm = Math.atan2(Pvd,Psd) + (int)(sm/Math.PI)*(2.*Math.PI);
	double Asp = Math.atan2(Pvt,Pst) + (int)(sp/Math.PI)*(2.*Math.PI);
	double Ast = Asp - Asm;
	tz[it] = (z[It] - z[Id])/Ast;

	firstRecal = false;
	return;
    }


    /**  If there are too many tracks, reject the worst ones. */
    protected void trimTracks(double radiusMin)
    {
	if (firstTrim) System.out.println(ClassName+"::trimTracks: ");

	int cutN=5;
	int ntimes=0;
	double ccCut = ccCuts;

	while (ntra >= MaxTracks) {
	    // ccCut *= 0.8;
	    int lastIt = ntra-1;
	    int minN=144;
	    if (firstTrkCut) System.out.println("  cut tracks w/ less than "+cutN+" points");
	    for (int it1=ntras; it1<ntra; it1++) {
		if (itra[it1][0] < minN) minN = itra[it1][0];
		if ((itra[it1][0] == cutN) &&
		    (ChiSq[it1]*radiusMin/80.) > ccCut) {
		    if (itra[it1][0] <= cutN) {
			int ip;
			for( ip=0; ip<= itra[lastIt][0]; ip++)
			    itra[it1][ip] = itra[lastIt][ip];
			ChiSq[it1] = ChiSq[lastIt];
			for(int i=0; i<5; i++)
			    trackPar[it1][i] = trackPar[lastIt][i];
			for (ip=1; ip<=itra[lastIt][0]; ip++)
			    path[it1][ip-1] = path[lastIt][ip-1];
			xd[it1] = xd[lastIt];
			yd[it1] = yd[lastIt];
			zd[it1] = zd[lastIt];
			xc[it1] = xc[lastIt];
			yc[it1] = yc[lastIt];
			xxc[it1] = xxc[lastIt];
			yyc[it1] = yyc[lastIt];
			si[it1] = si[lastIt];
			tz[it1] = tz[lastIt];
			outp[it1] = outp[lastIt];
			outt[it1] = outt[lastIt];
			ntra--;
			lastIt--;
		    }
		}
	    }
	    // if (minN < cutN) {
	    if (minN > cutN) {
		cutN = minN;
		// ccCut = ccCuts;
	    }
	    ntimes++;
	    if (ntimes>5) {
		System.out.println("  Quitting track cutting w/ cutN = "+cutN);
		return;
	    }
	    firstTrkCut = false;
	}
    }


    /** Set Monte Carlo particle parent. */
    void setMCParticle(int it, MCParticle mcp)
    {
        mcparticle[it] = mcp;
    }
    /** Put fraction of hits from parent MC particle. */
    void putMCPFraction(int it, double fraction)
    {
	mcpFraction[it] = fraction;
    }
    /** Put fraction of Tracker hits from parent MC particle. */
    void putMCPTrackerFraction(int it, double fraction)
    {
	mcpTrackerFraction[it] = fraction;
    }


    /** Sets tracker histogram options */
    protected void setHist()  { Hist = Hist || Hstd || (Hcand||Htrks||Hpnts||Hmiss); }

    /** Turns on Track finder production histograms. */
    public void setHprod() { Hstd = true; setHist(); setHtrks(); }
    /** Turns on Track finder test histograms. */
    public void setHtest() { setHprod(); setHcand(); setHtrks(); }
    /** Turns on Track finder development histograms. */
    public void setHdev() { setHprod(); setHcand(); }
    /** Turns on Track finder tuning histograms. */
    public void setHtune() { setHprod(); setHcand(); setHpnts(); }

    /** Turns on Track finder track candidate histograms: passing (curv,dz0,dzz,...) */
    public void setHcand() { Hcand = true; HEvents = 0; setHist(); }
    /** Turns on Track finder track histograms: track parameters (d0,phi0,...) */
    public void setHtrks() { Htrks = true; HEvents = 0; setHist(); }
    /** Turns on Track finder point histograms: point assignment (xy dev.,z res.,...) */
    public void setHpnts() { Hpnts = true; HEvents = 0; setHist(); }

    /** Turns off all histogram packages */
    public void setNoHist() { Hist = Hstd = Hcand = Htrks = Hpnts = Hmiss = false; }


    String ClassName = "TrackFinder::"; protected String text;
    protected int nEvt = 0;
    protected int NPrintEvents,NDebugEvents;

    // Errors
    protected boolean reportClosestApproachErrors = true;  protected int NReportClosestApproachErrors = 5;

    // Inline debugging and analysis
    protected boolean debug = false;  protected int MaxDebugEvts = 5;
    protected boolean trace = false;

    // Printing
    protected boolean first = true;
    protected boolean firstEvents = true;

    protected boolean firstPoints = false;
    protected boolean firstTrack  = true;
    protected boolean firstExtra  = false;
    protected boolean firstVtxPt  = false;
    protected boolean firstAmbigu = false;
    protected boolean firstRecal  = false;
    protected boolean firstTrim   = false;
    protected boolean firstTrkQC  = false;
    protected boolean firstTrkCut = false;

    protected boolean firstOrigin = false;
    protected boolean firstLooper = false;

    protected boolean firstFailWC   = false;
    protected boolean firstFailMaxP = false;
    protected boolean firstFailBW   = false;

    protected boolean allEvents   = false;
    protected boolean allLevels   = true;   // Set to allow finding first track at any level.
    protected boolean allTracks   = false;
    protected boolean allPoints   = false;

    protected DecimalFormat df = new DecimalFormat();

    // Histogramming
    protected String defaultFolder;
    protected int HEvents = 0, MaxHistEvts = 1000, MaxPntHistEvts = 1, MaxCandHistEvts = 10, MaxTrkHistEvts = 100;
    protected boolean Hist;
    protected boolean Hstd = false;
    protected boolean Hcand = false;
    protected boolean Htrks = false;
    protected boolean Hpnts = false;
    protected boolean Hmiss = false;

    Histogram HChiSqN;


    protected int ntra;
    protected int ntras = 0;
    protected double trackPar[][];    // Track parameters (d0,phi0,omega,z0,tandip)
    double minAz = 0.;
    double maxAz = 0.;
    protected int itra[][];           // Points on each track
    protected double xd[],yd[],zd[];  // ...
    protected double xc[],yc[];
    protected double xxc[],yyc[];
    protected double si[],tz[];
    protected boolean outp[],outt[];
    protected boolean fitsuccess[];

    protected double ChiSq[];       // ...
    protected double NDOF[];

    protected double ErrorMatrix[][][];

    protected MCParticle mcparticle[];

    protected double mcpFraction[];
    protected double mcpTrackerFraction[];

    /* Track finding strategy. */
    protected int MaxLevel;
    protected int NLevels;
    protected double minRadius[];
    protected int NPatterns;
    protected int triplets[][];
    protected int nPtsReq[];

    protected int Level = -1;
    protected int Pattern;

    protected int Paternmb = 0;
    protected int npchkd = 0;

    protected int tooMany = 20000000; // Max. # combinations 2 10^7

    /* Tracker specifications */
    protected double InnerRadius,OuterRadius;
    protected double HalfLength;
    protected int NLayers;
    protected double RPhiResolution,ZResolution;
    protected int NSigmaTracking;

    protected int Id,Im,It, Ibal;

    protected double dmas;            // phi tolerance (cm)
    protected double Dzas;            // z tolerance (cm)
    protected double rTol;            // distance from origin in xy
    protected double zTol;            // distance from origin in z
    protected double rPipe;           // beam pipe radius
    protected double rOut;            // solenoid radius

    /* Constants for Pat2 stratagy */
    protected double small = 0.000001;       // Min. size of triangular area (cm**2)
    protected double ccCuts = 30.;         // Used to eliminate poor tracks

    protected double ChiSq_factor = 0.4;

    protected Specifications specifications;
    //protected boolean VXDIncluded = false;

    protected TrackingStrategy strategy;

    protected Tracker tracker;
    protected TrackFitter fitter;
    protected VertexDetector vxd;
// EXTRA
    protected double Ddzax=0.;

    protected Histogram Hchis;

    protected boolean Heff = false;

    protected double path[][];
}
