/* CombinedTracker.java

		Reconstruct tracks in a combined CCD + TPC tracker

 Created by  Mike Ronan  Nov '98   Pick up LCD events within JAS framework and
                                   pass 3D points to pattern recognition software.
 Modified by:
   Nick Sinev
   Mike Ronan    May '01    Re-establish Tracking object model

 */

package hep.lcd.recon.ztracking.combined;

import hep.lcd.event.*;
import hep.lcd.recon.*;
import hep.lcd.recon.ztracking.*;
import hep.lcd.recon.ztracking.tpc.TPCSpecifications;
import hep.lcd.util.driver.*;

import hep.analysis.*;
import hep.physics.*;

import java.io.Serializable;
import java.util.*;
//import java.text.*;


/**   Picks up LCD events within JAS framework and passes 3D points to pattern recognition software */
final public class CombinedTracker extends AbstractTracker
{
    static int MAXPoints=25000;
    int MaxTracks;  // Obtained from TrackFinder

    int NVXDPnts;
    int[] nvxdpnt;


    public CombinedTracker()
    {
   	this(false);
    }
    /**  Create a combined tracking reconstructor using CombinedPat2 pattern recognition
     * and default specifications.
     */
    public CombinedTracker(boolean hist)
    {
	this(MAXPoints, new TPCSpecifications(), new CombinedPat2(hist), hist);
    }
    public CombinedTracker(int MaxPoints, Specifications specs, TrackFinder finder, boolean hist)
    {
	super(MaxPoints,specs,finder,hist);

	text = ClassName+"  version 1.1.8.7";
	text += "\n Combined track reconstruction using Pat2 - ";
	System.out.println(text);

	SimulateMeasurementError = false;
	VXDIncluded = true;

	MaxTracks = finder.getMaxTracks();
        nvxdpnt = new int[MaxTracks];

	if (debug) {
	  System.out.println("  Max. # points = "+MaxPoints);
	  if (!SimulateMeasurementError)
	      System.out.println("  Note: No measurement resolution smearing in combined detector reconstruction.");
	}

        setupParameters();
    }

    /** Get type of track hits. */
    public String getType() { return "Combined"; }


    /** Get number of vertex points on a track. */
    public int getNVXDPoints(int itrk)
    {
	return nvxdpnt[itrk];
    }
    /** Get number of tracker points on a track. */
    public int getNTrkPoints(int itrk)
    {
	return getNPoints(itrk)-nvxdpnt[itrk];
    }


    /** Get a space point on the track. */
    public double[] getTrackPoint(int itrk, int ipnt)
    {
	//if (debug && itrk==0 && ipnt<5) System.out.println(ClassName+"getTrackPoint: Get point #"+ipnt+" for track #"+itrk);
	int ntrk = getNTracks();
	if (itrk > ntrk) {
	    System.err.println(ClassName+"getTrackPoint: Requested track "+itrk+" does not exist. "+
			       "Only "+ntrk+" tracks available");
	    return null;  
	}
	int npnts = getNTrkPoints(itrk);
	if (ipnt > npnts) {
	    System.err.println(ClassName+"getTrackPoint: Requested point "+ipnt+" does not exist. "+
			       "Only "+npnts+" points available");
	    return null;  
	}
	int itra[] = getTrackPoints(itrk);
	int ip = itra[ipnt+nvxdpnt[itrk]];
	if (ip >= MaxPoints) {
	    System.err.println(ClassName+"getTrackPoint: *** Error in accessing tracker point "+ip+" ***");
	    return null;
	}
	double point[] = {x[ip],y[ip],z[ip]};
	return point;
    }

    /** this ID is unique for any point in tracking system  */
    public int getTrackHitID(int itrk,int ipnt)
    {
	//if (debug && itrk==0 && ip<5) System.out.println(ClassName+"getTrackHitID: Get point #"+ipnt+" for track #"+itrk);
	int ntrk = getNTracks();
	if(itrk > ntrk) {
	  System.err.println(ClassName+" Requested track "+itrk+" does not exist. "+
			     "Only "+ntrk+" tracks available.");
	  return -1;  
	}
	int npnts = getNTrkPoints(itrk);
	if(ipnt > npnts) {
	  System.err.println(ClassName+" Requested point "+ipnt+" does not exist. "+
			     "Only "+npnts+" points available."); 
	  return -1;  
	}
	int itra[] = getTrackPoints(itrk);
	int ip = itra[ipnt+nvxdpnt[itrk]];
	return 2*ip + 1;
    }

    /** Get the tracker hit for particular point on a track. */
    public TrackerHit getTrackHit(int itrk,int ipnt)
    {
	//if (debug && itrk==0 && ipnt<5) System.out.println(ClassName+"getTrackHit: Get point #"+ipnt+" for track #"+itrk);
	int ntrk = getNTracks();
	if (itrk > ntrk) {
	    System.err.println(ClassName+"getTrackHit: Requested track "+itrk+" does not exist. "+
			       "Only "+ntrk+" tracks available.");
	    return null;  
	}
	int npnts = getNTrkPoints(itrk);
	if (ipnt > npnts) {
	    System.err.println(ClassName+"getTrackHit: Requested point "+ipnt+" does not exist. "+
			       "Only "+npnts+" points available."); 
	    return null;  
	}
	int itra[] = getTrackPoints(itrk);
	int ip = itra[ipnt+nvxdpnt[itrk]];
	if (trkHit[ip]==null)
	    System.err.println(ClassName+"getTrackHit: Null TrackerHit ???");
	return trkHit[ip];
    }


    public void process(LCDEvent event)
    {
	nEvt++;
	if (firstEvents) {
	  text = ClassName+" Find tracks in combined detector \t\t Event #"+nEvt;
	  System.out.println(text);
	  if (batch) System.err.println(text);
	}

	/** Update Tracker and Vertex Detector specifications. */
	if (first) System.out.println(" Update specifications if needed.");
	specifications.update(event);
	if (specifications.areNew()) {
	  setupParameters();
	  if (vxd!=null) vxdSpecifications.update(event);
	  setNLayers();
	}

	// Set up track fitter
	//if (fitter != null) fitter.setupEvent(event);
	if (fitter != null) {}
	//else if (first) System.out.println(ClassName+" Warning: Fitter is not set!");

	if (first) {
	  text = " NLayers = "+NLayers;
	  if (vxd!=null) text+=" including "+NVXDLayers+" vertex detector layers.";
	  System.out.println(text);
	  System.out.println(" Tracker specifications: ");
	  System.out.println("  \t Inner,Outer radii = "+specifications.getInnerRadius()+","+specifications.getOuterRadius());
	  System.out.println("  \t Half-length = "+specifications.getHalfLength());
	  System.out.println("  \t B Field     = "+specifications.getBField());
	  if (vxd!=null) {
	    System.out.println(" Vertex Detector specifications: ");
	    System.out.println("  \t Inner,Outer radii = "+vxdSpecifications.getInnerRadius()+","+vxdSpecifications.getOuterRadius());
	    System.out.println("  \t Half-length = "+vxdSpecifications.getHalfLength());
	    System.out.println("  \t B Field     = "+vxdSpecifications.getBField());
	  }
	}


	/** Get tracker hits. */
	for(int i=0; i<MaxTracks; i++)
	    nvxdpnt[i] = 0;

	TrackerHits vxdhits = event.getVXDHits();
	TrackerHits trkhits = event.getTrackerHits();

	if (Hist) {
	    HEvents++;
	    HistogramFolder.setDefaultFolder(defaultFolder);
	    if (Hevts && HEvents<=MaxHistEvts) {
		histogram("VXD NHits/evt").fill(vxdhits.getNHits());
		histogram("Trk NHits/evt").fill(trkhits.getNHits());
	    }
	}


	/** Determine number of hits in each layer of tracking detectors. */
	int ncl=0;
	int nlayer[] = new int[NLayers];
	for (int i=0;i<NLayers;i++) { nlayer[i]=0; }

	// Count number of hits on each Vertex Detector layer.
	int nVXD = 0;
	if (vxd!=null) {
	  if (first) System.out.println(" Count # of Vertex Detector hits on each layer.");
	  for (Enumeration hits = vxdhits.getHits();(hits.hasMoreElements() && ncl<MaxPoints);) {
	    TrackerHit hit = (TrackerHit) hits.nextElement();
	    /// int layer = hit.getLayer();
	    /// if(layer > 127) layer-=128;
	    //int layer = hit.getLayer() & 0x00000007;
	    /// double specr=vxdSpecifications.getLayerRadius(layer);
	    /// double point[] = hit.getPoint();
	    /// double acrad=Math.sqrt(point[0]*point[0]+point[1]*point[1]);
	    /// if (Math.abs(specr-acrad) < 0.5) {
	    int layer = hit.getLayer();
	    if (Hhits && HEvents<=MaxHitHistEvts) histogram("Hits/layer").fill(layer);
	    nlayer[layer]++;
	    ncl++;
	  }
	  nVXD = ncl;
	  if (first) {
	    //text = "  VXD hits/layer counted, # VXD hits "+nVXD;
	    text = "  Total of "+nVXD+" Vertex Detector hits found.";
	    System.out.println(text); if (batch) System.err.println(text);
	  }
	}

	// Count number of hits on each Tracker layer.
	if (first) System.out.println(" Count # of Tracker hits on each layer.");
	for (Enumeration hits = trkhits.getHits();(hits.hasMoreElements() && ncl<MaxPoints);) {
	    TrackerHit hit = (TrackerHit) hits.nextElement();
	    int layer = hit.getLayer() + NVXDLayers;
	    /// double point[] = hit.getPoint();
	    /// double specr=specifications.getLayerRadius(tlayer);
	    /// double acrad=Math.sqrt(point[0]*point[0]+point[1]*point[1]);
	    /// if (Math.abs(specr-acrad) < 0.5) {
	    if (Hhits && HEvents<=MaxHitHistEvts) histogram("Hits/layer").fill(layer);
	    nlayer[layer]++;
	    ncl++;
	}
	//if (printPoints && ncl>10) System.out.println(" ...");

	int nTrk = ncl - nVXD;
	if (first) {
	  //text = "  Tracker hits/layer counted, # Trk hits "+nTrk;
	  text = "  Total of "+nTrk+" Tracker hits found.";
	  System.out.println(text); if (batch) System.err.println(text);
	}


	/** Get hits from each tracking detector, ordered by layer number. */
	int nPoints=0;

	ipoi[0] = ilon[0] = 0;
	for (int layer=1; layer<NLayers; layer++) {
	    ipoi[layer] = ipoi[layer-1]+nlayer[layer-1];
	    ilon[layer] = 0;
	}
	NVXDPnts = ipoi[NVXDLayers];

	if (vxd!=null) {
	  if (SimulateMeasurementError) {
	    double RPhiResolution = vxdSpecifications.getRPhiResolution();
	    double ZResolution    = vxdSpecifications.getZResolution();

	    if (first) {
	      System.out.println(" Get Vertex Detector points and simulate measurement errors.");
	      System.out.println("  Measurement resolutions: rphi = "+RPhiResolution+", z = "+ZResolution);
	    }
	    for (Enumeration hits = vxdhits.getHits();(hits.hasMoreElements() && nPoints<MaxPoints);) {
	      TrackerHit hit = (TrackerHit) hits.nextElement();
	      int layer = hit.getLayer();
	      double point[] = hit.getPoint();
	      int iptr = ipoi[layer]+ilon[layer];
	      double r = Math.sqrt(point[0]*point[0] + point[1]*point[1]);
	      double phi = Math.atan2(point[1],point[0]);
	      double RPhi = r*phi + random.nextGaussian()*RPhiResolution;
	      phi = RPhi / r;
	      x[iptr] = r * Math.cos(phi);  y[iptr] = r * Math.sin(phi);
	      z[iptr] = point[2] + random.nextGaussian()*ZResolution;
	      KFlag[iptr] = false;
	      trkHit[iptr] = hit;
	      ilon[layer]++;
	      nPoints++;
	    }

	  } else {
	    if (first)
	      System.out.println(" Note: No Vertex Detector measurement resolution smearing in track reconstruction.");

	    if (first) System.out.println(" Get Vertex Detector points");
	    for (Enumeration hits = vxdhits.getHits();(hits.hasMoreElements() && nPoints<MaxPoints);) {
	      TrackerHit hit = (TrackerHit) hits.nextElement();
	      int layer = hit.getLayer();
	      double point[] = hit.getPoint();
	      int iptr = ipoi[layer]+ilon[layer];
	      x[iptr] = point[0];  y[iptr] = point[1];  z[iptr] = point[2];
	      KFlag[iptr] = false;
	      trkHit[iptr] = hit;
	      ilon[layer]++;
	      nPoints++;
	    }
	  }
	} // vxd!=null
	if (first) System.out.println("  Got "+nPoints+" Vertex Detector points.");

	if (Hhits && HEvents<=MaxHitHistEvts) {
	  for (int i=0;i<NVXDPnts;i++) {
	    if (HEvents<=1) {
		histogram("VXD y vs.x").fill(x[i],y[i]);
		double r = Math.sqrt(x[i]*x[i]+y[i]*y[i]);
		histogram("VXD r vs.z").fill(z[i],r);
	    }
	    histogram("VXD z").fill(z[i]);
	    double phi = Math.atan2(y[i],x[i]);
	    histogram("VXD phi").fill(phi);
	    //if ((0.<phi && phi<1.) && z[i]>0.);
	    //histogram("VXD wedge r vs.z").fill(z[i],r);
	  }
	}


	if (SimulateMeasurementError) {
	  double RPhiResolution = specifications.getRPhiResolution();
	  double ZResolution    = specifications.getZResolution();

	  if (first) {
	      System.out.println(" Get Tracker points and simulate measurement errors.");
	      System.out.println("  Measurement resolutions: rphi = "+RPhiResolution+", z = "+ZResolution);
	  }
	  for (Enumeration hits = trkhits.getHits();(hits.hasMoreElements() && nPoints<MaxPoints);) {
	    TrackerHit hit = (TrackerHit) hits.nextElement();
	    int layer = NVXDLayers + hit.getLayer();
	    double point[] = hit.getPoint();
	    int iptr = ipoi[layer]+ilon[layer];
	    double r = Math.sqrt(point[0]*point[0] + point[1]*point[1]);
	    double phi = Math.atan2(point[1],point[0]);
	    double RPhi = r*phi + random.nextGaussian()*RPhiResolution;
	    phi = RPhi / r;
	    x[iptr] = r * Math.cos(phi);  y[iptr] = r * Math.sin(phi);
	    z[iptr] = point[2] + random.nextGaussian()*ZResolution;
	    KFlag[iptr] = false;
	    trkHit[iptr] = hit;
	    ilon[layer]++;
	    nPoints++;
	  }

	} else {
	  if (first)
	      System.out.println("  Note: No Tracker measurement resolution smearing in track reconstruction.");

	  if (first) System.out.println(" Get Tracker points");
	  for (Enumeration hits = trkhits.getHits();(hits.hasMoreElements() && nPoints<MaxPoints);) {
	    TrackerHit hit = (TrackerHit) hits.nextElement();
	    int layer = NVXDLayers + hit.getLayer();
	    double point[] = hit.getPoint();
	    int iptr = ipoi[layer]+ilon[layer];
	    x[iptr] = point[0];  y[iptr] = point[1];  z[iptr] = point[2];
	    KFlag[iptr] = false;
	    trkHit[iptr] = hit;
	    ilon[layer]++;
	    nPoints++;
	  }
	}
	if (first) System.out.println("  Total of "+nPoints+" points found.");


	if (Hhits && HEvents<=MaxHitHistEvts) {
	  for (int i=NVXDPnts;i<nPoints;i++) {
	    if (HEvents<=1) {
		histogram("Trk y vs.x").fill(x[i],y[i]);
		double r = Math.sqrt(x[i]*x[i]+y[i]*y[i]);
		histogram("Trk r vs.z").fill(z[i],r);
	    }
	    histogram("Trk z").fill(z[i]);
	    double phi = Math.atan2(y[i],x[i]);
	    histogram("Trk phi").fill(phi);
	    //if ((0.<phi && phi<1.) && z[i]>0.)
	    //histogram("Trk wedge r vs.z").fill(z[i],r);
	  }
	}


	// ???
	if (vxd!=null) vxd.startEvent(event);

	/** Find tracks. */
	if (first) System.out.println(" Find tracks at "+NLevels+" levels.");
	for (int level=MinLevel; level<NLevels; level++) {
	  if (first) System.out.println(" Find tracks at level "+level);

	  //finder.findTracks(level,nPoints,x,y,z,KFlag,specs,NLayers,ipoi,ilon);
	  finder.findTracks(level,nPoints,NLayers);

	  if (first)
	    System.out.println("  At level "+level+", number of tracks found = "+getNTracks());
	}

	// Get number of tracks found
	NTracks = getNTracks();


	// Determine number of vertex points on each track
	if (vxd!=null) {
	  for (int i=0;i<NTracks;i++) {
	    nvxdpnt[i] = 0;
	    int itra[] = getTrackPoints(i);
	    int ntp = getNPoints(i);
	    for (int j=1; j<=ntp; j++)
		if (itra[j]<NVXDPnts) nvxdpnt[i]++;
	  }
	  vxd.finishEvent();
	}

	if (first) {
	  if (vxd!=null) {
	    System.out.println(" Number of vertex points on tracks: ");
	    for (int it=0;it<NTracks;it++) {
	      int ntp = getNPoints(it);
	      if (it<NTrksPrint)
		System.out.println("  Track "+it+" # points "+ntp+" w/ "+nvxdpnt[it]+" VXD points");
	    }
	    if (NTracks>=NTrksPrint) System.out.println("  ...");
	  }
	}


	/** Provide standard histograms. */
	if (Hist) {
	  if (Hstd||Htrks) {
	    histogram("NTracks").fill(NTracks);
	    if (HEvents<MaxTrkHistEvts) {
	      for (int it=0;it<NTracks;it++) {
		//histogram("NPoints/trk").fill(getNPoints(it));
		if (vxd!=null)
		  histogram("# VXD points/trk").fill(getNVXDPoints(it));
		histogram("# Trk points/trk").fill(getNTrkPoints(it));
		if (Htrks) {
	          HistogramFolder.setDefaultFolder("Tracks");
		  histogram("d0").fill(get_d0(it));
		  histogram("phi0").fill(get_phi0(it));
		  histogram("curv").fill(get_curv(it));
		  histogram("tandip").fill(get_tandip(it));
		}
		histogram("z0").fill(get_z0(it));
		if (Htrks)
		    HistogramFolder.setDefaultFolder("..");
	      }
	    }
	  }
	  if (Hist) HistogramFolder.setDefaultFolder("/");
	}

	if (firstEvents && summarize) {
	    text = "";
	    if (first) text = "Combined tracking ("+getType()+") reconstruction summary: \n";
	    text += "  Number of tracks found = "+NTracks;
	    System.out.println(text);  if (printDebug) out.println(text);
	    if (batch) System.err.println(text);
	}

	if (first) first = printPoints = false;
	if (nEvt>=NPrintEvents)
	    firstEvents = printDebug = false;

    } // End process


    String ClassName = "CombinedTracker::";
}
