/* WeightMatrix.java

  This class constructs measurement weight matrix to fit tracks with tracker
  hits (both CCD and TPC)
  To construct such matrix this class needs Track or TrkParams as argument, and
  it extracts detector properies from detector.ini file, and tracking device
  resolutions from HitsSmearing.ini files
  The weight matrix is invers of squares of measurement errors matrix. Measurement errors
  are generated by detector resolutions and multiple scattering. 
  For now we will assume
  that errors  due to detector resolution are independent for all points, though it is 
  possible to introduce correlated errors due to TPC superlayers position error, for example.
  For the N measurement points we will consider having 2*N measurements - position of each
  point in XY plane and Z coordinate. This corresponds to 2*N residuals, which we can calculate
  (there is no residual in the radius of the hit - we assume we know radius of every hit exactly)
  So weight matrix will have (2*N)*(2*N) elements.In fact it may be (2*(N+1))**2, because we may
  add beam pipe scattering to errors, though there is no measurements in "Beam Pipe" layer.
  Measurement error squre for each measurement is sum of detector resolution and multiple
  scattering term. Multipple scattering for diagonal elements are calculated as summ of all 
  deviations from scattering in all layers preceeding given one.
  If some layer K generates multiple scattering angle THETk, then linear displacement at the
  exit of this layer is 1/sqrt(3) * dLk * THETk. Here dLk is the particle path length in layer k.
  The displacement in the layer I caused by MS in layer K will be: (Li - Lk) * THETk + mentioned
  above 1/sqrt(3) * dLk * THETk, so total (Li-Lk+1/sqrt(3)*dLk)*THETk, and because contributions
  from different layers are independent, they will add in square, so

  SIGii**2 = SUMk[((Li-Lk+1/sqrt(3)*dLk)*THETk)**2]

  For non-diagonal elements

  SIGij = SUMk[(Li-Lk+dLk/sqrt(3))*THETk*(Lj-Lk+dLk/sqrt(3)*THETk] 

  (I follow here prescription  from MOURS Benoit, who created fitting routines for SLD)

  During track fitting we will calculate weight matrix for each track only once, assuming that it
  does not change considerably due to track parameters change during the fit.

  For tracker with large number of layers, we will group 12 layers of tracker (not vertex detector)
  into 1 point. Measurement errors for such point is assumed decreased by sqrt 12, and multiple
  scaterring is calculated at the center of the group.

  Author Nick Sinev (650)-926-2913


 Created by  Nick Sinev  Mar '99   
 Modified by
   Mike Ronan    May '01      - 

 */

package hep.lcd.recon.ztracking.fitters.sld;

import hep.lcd.event.*;
import hep.lcd.geometry.*;
import hep.lcd.geometry.component.Field_Strength;
import hep.lcd.recon.ztracking.*;

import Jama.*;

import java.text.*;
import java.util.*;


public class WeightMatrix
{
    final int MAXLAYERS = 160;

    Detector d;
    Track tr;
    TrackParameters tpar = new TrackParameters(0.,0.,0.,0.,0.);
    Matrix wm;
    HelicalSwimmer swmr = new HelicalSwimmer();
    double[] r0 = new double[3];
    double[] p0 = new double[3];
    int tchg;
    double[] pathlen;
    double[] lradlen;
    double[] lradii;
    boolean[] active; 
    int npnts=0;

    boolean useBP = false; 
    boolean useTracker = false;
    boolean useCCD = false;

    int ngrp = 1;
    int itpcr = 0;
    double halfpi = 0.5 * Math.PI;
    double xanc[] = new double[MAXLAYERS];
    TrackingMaterials materials = new TrackingMaterials();


    public WeightMatrix()
    {
    }


    boolean check;

    public boolean putTrack(Track track, boolean debug)
    {
	check = false;

	dbg = debug;
	d = Detector.instance();
	Field_Strength field = (Field_Strength) d.get("Field_Strength");
	double BField = field.getField();
	this.tr = track;
	for (int i=0; i<3; i++) {
	    p0[i] = tr.getMomentum()[i];
	    r0[i] = tr.getOrigin()[i];
	}
	tchg = tr.getCharge();
	double pt = Math.sqrt(p0[0]*p0[0] + p0[1]*p0[1]);
	double ph0 = Math.atan2(p0[1],p0[0]);
	tpar.put_phi0(ph0);
	double phor = ph0 + tchg * halfpi;
	if (Math.abs(Math.cos(phor)) > Math.abs(Math.sin(phor)))
	    tpar.put_d0(r0[0]/Math.cos(phor));
	else
	    tpar.put_d0(r0[1]/Math.sin(phor));
	tpar.put_omega(tchg * BField*0.003/pt);
	tpar.put_z0(r0[2]);
	tpar.put_s(p0[2]/pt);

	doWeightMatrix();  

	return check;
    }
 
    public void putParameters(TrackParameters tp)
    {
	if (dbg) System.out.println(" Start weight matrix calc.");
	d = Detector.instance();
	this.tpar = tp;
	doWeightMatrix();
    }

    void doWeightMatrix()
    {
	df.setMaximumFractionDigits(5);
	if (dbg) System.out.println(" Calculating weight matrix");
    
	PropertySet tpc = (PropertySet) d.get("MultiLayer_TRACK_Barrel");
	useTracker = false;
	if(tpc != null) {
	    useTracker = true;
	    //if (dbg) System.out.println("  Processing TRACKER");
	    double nob = tpc.getDouble("num_of_barrels");
	    int inb = (int) nob;
	    int nl = 0;
	    int nlt = 0;
	    for (int i = 1; i<inb+1; i++) {
		double nlpb = tpc.getDouble("layers"+i);
		int inlpb = (int) nlpb;
		nlt+= inlpb;
	    }
	    ngrp = 1;
	    if (nlt > 36) ngrp = nlt/12;
	    //if (dbg) { text = "   Number of barrels: "+inb+"\n     Total number of layers: "+nlt+"\n     Number of layers in each TRACKER group is "+ngrp; System.out.println(text); }
	}
	double r_cyl = 100.;
	double z_pl = 1000.;
	swmr.swim(tpar,r_cyl,z_pl);
	materials.setGroup(ngrp);
	materials.putParameters(tpar);
	double[] lengths = materials.getLengths();
	lradii = materials.getRadii();
	active = materials.getActive();
	npnts=0;
	for (int i=0; i<materials.getNLayers(); i++)
	    if (active[i]) npnts++;

	// Require min. number of points
	if (npnts==0) return;

	useBP = true;
	itpcr = 0;    
	PropertySet vtx = (PropertySet) d.get("MultiLayer_VERTX_Barrel");
	if (vtx != null) {
	    useCCD = true;
	    //if (dbg) System.out.println("  Processing VXD"); 
	    double nob = vtx.getDouble("num_of_barrels");
	    int inb = (int) nob;
	    int nl = 0;
	    for (int i = 1; i<inb+1; i++) {
		double innr = vtx.getDouble("inner_radius"+i);
		z_pl = vtx.getDouble("z_outer"+i);
		double nlpb = vtx.getDouble("layers"+i);
		int inlpb = (int) nlpb;
		//if (dbg) System.out.println("   barrel "+i+" layers "+inlpb+" in.r "+innr);
		nl+=inlpb;
	    }
	    itpcr = nl;
	}
	double ph0 = tpar.get_phi0();
	double d0 = tpar.get_d0();
	Field_Strength field = (Field_Strength) d.get("Field_Strength");
	double BField = field.getField();
	//if (dbg) System.out.println("  Field = "+BField);
	double pt = BField*0.3*.01/Math.abs(tpar.get_omega());
	tchg = tpar.get_omega() > 0. ? +1 : -1;
	p0[0]=pt*Math.cos(ph0);
	p0[1]=pt*Math.sin(ph0);
	p0[2]=pt*tpar.get_tandip();
	double sint = pt/(Math.sqrt(p0[0]*p0[0]+p0[1]*p0[1]+p0[2]*p0[2]));
	pathlen = new double[npnts+1];
	for (int i=0; i<npnts; i++) {
	    if (active[i]) pathlen[i] = lengths[i];
	    else pathlen[i] = lradii[i]/sint;
	}
	pathlen[npnts]=pathlen[npnts-1];  // Need to protect against 0 points  
	lradlen = materials.getThick();
	//if (dbg) System.out.println("  Calculate Weight matrix");
	calculateWM();

	check = true;
    }


    void calculateWM()
    {
	/* we will use detector resolutions from HitsSmearing.ini files.
	 * REMEMBER - all resolutions, residuals and so on are in cm .
	 */
	double pim = 0.1396; // assume pion mass for calculating beta);
	double oosq3 = 0.57735;
	double merr[][] = new double[2*npnts][2*npnts];
	for (int i=0; i<2*npnts; i++)
	    for (int j=0; j<2*npnts; j++)
		merr[i][j] = 99.;  
	PropertySet parm = d.getParameters("HitSmearing");
	double txyer = parm.getDouble("TPCXYSmear");
	double tzer  = parm.getDouble("TPCZSmear");
	double vxyer = parm.getDouble("CCDXYSmear");
	double vzer  = parm.getDouble("CCDZSmear");
	double ptots = p0[0]*p0[0] + p0[1]*p0[1] + p0[2]*p0[2];
	double ptot = Math.sqrt(ptots);
	double ptrs = p0[0]*p0[0]+p0[1]*p0[1];
	double ptr = Math.sqrt(ptrs); 
	double oneovcls = ptots/ptrs; 
	double beta = ptot/Math.sqrt(pim*pim + ptot*ptot);
	double coef = 0.0136/(beta*ptot);
	double thetk[] = new double[npnts];
	df.setMaximumFractionDigits(5);
	if (useBP) itpcr++;
	for (int i=0; i<npnts; i++) {
	    swmr.swim(lradii[i]);
	    double sinxs = swmr.getSinXang();
	    xanc[i] = 1.0;  
	    if (sinxs < 0.1) xanc[i] = 100.;
	    if (sinxs > 0.09) xanc[i] = 1./(sinxs * sinxs);
	    if (xanc[i] < 1.)
		System.out.println("Cros. angle corr. for point "+i+" : "+xanc[i]);
	    if (xanc[i] > 100.) xanc[i] = 100.;
	    double eln = lradlen[i];
	    double lcor = 1. + 0.038 * Math.log(eln);
	    thetk[i] = coef * lcor * Math.sqrt(eln);
	}

	double grp = (double) ngrp;
	double grpf = Math.sqrt(grp);
	for (int i = 0; i < npnts; i++) {
	    double me = 0.;
	    for (int j = 0; j < i+1; j++) {
		if (i==j) {
		    me = (i<itpcr) ? vxyer : txyer ;
		    if (i>itpcr-1) me/=grpf;
		    //if (i<itpcr) me = me*100.;
		}
		double msersq = 0.;
		if (j > 0) {
		    for (int k = 0; k<j; k++) {
			double dLkt = (oosq3+0.5) * Math.abs((pathlen[k+1]-pathlen[k]));
			if (k==i) dLkt=0.;
			if (active[k]) {
			    double delpi = 0.;
			    double delpj = 0.;
			    if (i > k+1) delpi = pathlen[i]-pathlen[k+1];
			    if (j > k+1) delpj = pathlen[j]-pathlen[k+1];
			    msersq+= thetk[k]*thetk[k]*(delpi+dLkt)*(delpj+dLkt);
			}
		    }
		}
		msersq = msersq * xanc[j]; 
		if (j==i) merr[i][j] = me*me + msersq ;
		if (i != j) merr[i][j] = msersq ;
		if (i != j) merr[j][i] = merr[i][j];  
	    }
	}
	for (int i = npnts; i < 2*npnts; i++) {
	    double me = 0.;
	    for (int j = 0; j < i+1; j++) {
		int ie = i-npnts;
		int je = j-npnts;
		if (i==j) {
		    me = (ie<itpcr) ? vzer : tzer ;
		    if (i>itpcr-1) me/=grpf;
		    //if (i < itpcr) me = me*100.;
		}
		if (je<1) me = 0.;
		double msersq = 0.;
		if (je > 1) {       
		    for (int k = 1; k<je; k++) {
			double dLkt = (oosq3+0.5) * (pathlen[k+1]-pathlen[k]);
			if (k==ie) dLkt = 0.;
			double delpi = 0.;
			double delpj = 0.;
			if (ie > k+1) delpi = pathlen[ie]-pathlen[k+1];
			if (je > k+1) delpj = pathlen[je]-pathlen[k+1];
			if (active[k])
			    msersq+= thetk[k]*thetk[k]*(delpi+dLkt)*(delpj+dLkt);
		    }
		}
		msersq*=oneovcls;
		if (i==j) merr[i][j] = me*me + msersq;
		if (i != j) merr[i][j] = msersq ;
		if (i != j) merr[j][i] = merr[i][j];  
	    }
	}
	if (useBP) {
	    merr[0][0] = 999.;
	    merr[npnts][npnts] = 999.;
	}
/*
	for (int i=0; i<npnts; i++) {
	    double errinxy = merr[i][i] * 10000.;
	    double errinz = merr[i+npnts][i+npnts] * 10000.;
	    System.out.println("layer "+i+" RPHI err = "+df.format(errinxy)+" Z err = "+df.format(errinz));
	}
*/
	Matrix ers = new Matrix(merr);
	wm = ers.inverse();

	if (debug) {
	  System.out.println(ClassName+" Weight matrix calculated");
	  for (int i=0; i<npnts; i++) {
	    double wtxy = wm.get(i,i);
	    double wtz = wm.get(i+npnts,i+npnts);
	    // System.out.println("layer "+i+" RPHI wt = "+df.format(wtxy)+" Z wt = "+df.format(wtz));
	  }
	}
    }

    public Matrix getMatrix()
    {
	return wm;
    }

    public boolean[] getActive()
    {
	return active;
    }

    String ClassName = "WeightMatrix::"; String text;

    // Inline debugging and analysis
    boolean debug = false;
    boolean dbg = false;

    DecimalFormat df = new DecimalFormat();
}
